#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# using print() built infunction, disable print statement
from __future__ import print_function

#
# wxPython is not part of the standard Python distribution and has to be downloaded and installed separately.
# Tell the user that wxPython is required but has not been found
#
try:
  import wx
except ImportError:
  raise ImportError, "Please download the appropriate version of wxPython from www.wxpython.org"

try:
  import os
  import shutil
except ImportError:
  raise ImportError, "import OS failed"

from defines import DEFINES
from pdef import MLELEGACY_DEF
from ElementGui import *

from util import UTILS
utilities = UTILS()

try:
    import cPickle as pickle
except ImportError:
    import pickle         # fall back on Python version

# TXT Policy Generator Tool
# MLE Class - Policy Definition File Lists
#
class MLELegacy( ElementGui ):

  CONST_TITLE = "Choose Hash File"
  CONST_WILDCARD = "Hash file (*.hash) | *.hash|" \
                   "All Files (*.*)    | *.*"

  """__init__() - MLE class constructor"""
  def __init__( self ):
    self.mlePanelWidgets = []
    self.panelCreated = False

  #
  # create the MLE Panel
  #
  def createOrShowPanel(self, wx, listPanel, parent, pdef, statusBar):
    """createMlePanel - create the List Panel"""

    #print("createOrShowPanel panelCreated == %s" % (self.panelCreated))    # DBGDBG
    # 1st time, create the panel
    # nth time, show the panel
    if(self.panelCreated == True):
      self.showPanel()
      return

    self.pdef = pdef
    self.parent = parent
    self.listPanel = listPanel
    self.StatusBar = statusBar
    parentSizer = parent.GetSizer()

    currentList = self.pdef.getCurrentListObject()
    self.myIndex = len(currentList.ElementDefData)-1    # Just added the element, the last one should be the one.

    # create the Mle Panel sizers
    #self.mlePanelSizer = wx.BoxSizer(wx.VERTICAL)
    self.mleGridSizer= wx.GridBagSizer(hgap=5, vgap=5)
    #self.mleHorizSizer = wx.BoxSizer(wx.HORIZONTAL)

    self.mlePanel = wx.Panel(parent, -1)
    self.mlePanel.SetSizer(self.mleGridSizer)

    mleLabelText1 = "MLE"
    mleLabelText2 = "Element"
    mleLabel1 = wx.StaticText(self.mlePanel, -1, mleLabelText1)
    mleLabel2 = wx.StaticText(self.mlePanel, -1, mleLabelText2)
    font = wx.Font( 18, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    mleLabel1.SetFont( font )
    self.mleGridSizer.Add( mleLabel1, pos=(0, 3))
    self.mlePanelWidgets.append(mleLabel1)
    mleLabel2.SetFont( font )
    self.mleGridSizer.Add( mleLabel2, pos=(0, 4))
    self.mlePanelWidgets.append(mleLabel2)

    typeLabel = wx.StaticText(self.mlePanel, label="Type")
    self.mleGridSizer.Add( typeLabel, pos=(1,3))
    self.mlePanelWidgets.append(typeLabel)
    typeEdit  = wx.TextCtrl( self.mlePanel, value="MLE", size=(40, -1))
    typeEdit.Enable( False )
    self.mleGridSizer.Add( typeEdit,  pos=(1,4))
    self.mlePanelWidgets.append(typeEdit)

    contolOptionsLabel = wx.StaticText(self.mlePanel, -1, "Control")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    contolOptionsLabel.SetFont( font )

    # override PS policy bit is applicable only if PO policy rules
    self.overridePsPolicy = wx.CheckBox(self.mlePanel, label="Override PS Policy")

    if(self.pdef.Rules == DEFINES.PoRules):
      self.overridePsPolicy.Enable( True )
    else:
      self.overridePsPolicy.Enable( False )

    self.mleGridSizer.Add(contolOptionsLabel, pos=(0,9), span=(1,2), flag=wx.BOTTOM, border=5)
    self.mleGridSizer.Add(self.overridePsPolicy,   pos=(1,9), span=(1,2), flag=wx.BOTTOM, border=5)
    self.overridePsPolicy.Bind(wx.EVT_CHECKBOX, self.onOverridePsPolicy)
    self.mlePanelWidgets.append(contolOptionsLabel)
    self.mlePanelWidgets.append(self.overridePsPolicy)

    # STPM is required bit ElementPolicyControl[1]
    #self.stmIsRequired = wx.CheckBox(self.mlePanel, label="STM is required")
    #self.mleGridSizer.Add(self.stmIsRequired, pos=(2,9), span=(1,2), flag=wx.BOTTOM, border=5)
    #self.stmIsRequired.Bind(wx.EVT_CHECKBOX, self.onOverridePsPolicy)  # TODO: add function to handle event.
    #self.mlePanelWidgets.append(self.stmIsRequired)
    
    hashList = ['SHA1']
    hashAlgLabel = wx.StaticText(self.mlePanel, label="Hash Algorithm")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    hashAlgLabel.SetFont( font )
    self.mleGridSizer.Add(hashAlgLabel, pos=(0,20))
    self.mlePanelWidgets.append(hashAlgLabel)

    hashAlgEdit = wx.ComboBox( self.mlePanel, size=(75, -1), value="SHA1", choices=hashList, style=wx.CB_DROPDOWN )
    self.mleGridSizer.Add(hashAlgEdit, pos=(1,20))
    self.mlePanelWidgets.append(hashAlgEdit)

    minSinitVersionLabel = wx.StaticText(self.mlePanel, label="Min SINIT Version: ")
    self.mleGridSizer.Add(minSinitVersionLabel, pos=(2,4))
    self.mlePanelWidgets.append(minSinitVersionLabel)
    minSinitVersion = pdef.SinitMinVersion                               # get current value
    currentList.ElementDefData[self.myIndex].SinitMinVersion = minSinitVersion

    self.minSinitVersionEdit  = wx.TextCtrl( self.mlePanel, value=str(minSinitVersion), size=(30, -1))
    self.mleGridSizer.Add( self.minSinitVersionEdit,  pos=(2,5))
    self.minSinitVersionEdit.Bind(wx.EVT_TEXT, self.onMinSinitVersion)
    self.mlePanelWidgets.append(self.minSinitVersionEdit)

    self.addButton = wx.Button( self.mlePanel, -1,      label="    Add   ")
    self.mleGridSizer.Add( self.addButton, pos=(4,3))
    self.mlePanelWidgets.append(self.addButton)
    self.addButton.Bind(wx.EVT_BUTTON, self.onAddButtonClick)
    self.mlePanelWidgets.append(self.addButton)

    self.removeButton = wx.Button( self.mlePanel, -1,      label="  Remove  ")
    self.mleGridSizer.Add( self.removeButton, pos=(4,5))
    self.mlePanelWidgets.append(self.removeButton)
    self.removeButton.Bind(wx.EVT_BUTTON, self.onRemoveButtonClick)
    self.mlePanelWidgets.append(self.removeButton)

    self.hashListLabel = wx.StaticText(self.mlePanel, label="      Hash File List")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    self.hashListLabel.SetFont( font )
    self.mleGridSizer.Add( self.hashListLabel, pos=(3,4))
    self.mlePanelWidgets.append(self.hashListLabel)

    self.fileCntLabel = wx.StaticText(self.mlePanel, label="Number of Files")
    self.mleGridSizer.Add( self.fileCntLabel, pos=(5,3))
    self.mlePanelWidgets.append(self.fileCntLabel)
    self.fileCntEdit  = wx.TextCtrl( self.mlePanel, value="0", size=(40, -1))
    self.fileCntEdit.Enable( False )
    self.mleGridSizer.Add( self.fileCntEdit,  pos=(5,4))
    self.mlePanelWidgets.append(self.fileCntEdit)

    hashFileList = ['']

    self.hashListBox = wx.TextCtrl( self.mlePanel, value="", size=(150, 120), style = wx.TE_MULTILINE)  # Note: add |wx.HSCROLL to get a horiz scroll bar
    # hashListBox must be enabled so can select items to remove
    self.hashListBox.Bind(wx.EVT_TEXT, self.onHashListBoxEdit)
    self.hashListBox.SetInsertionPoint(0)
    self.mleGridSizer.Add( self.hashListBox, pos=(4,4))
    self.mlePanelWidgets.append(self.hashListBox)

    self.mlePanelWidgets.append(self.mlePanel)

    #self.mleHorizSizer.Add(self.mleGridSizer,  0, wx.ALL, 5)
    #self.mlePanelSizer.Add(self.mleHorizSizer, 0, wx.ALL, 5)
    #parent.SetSizerAndFit(self.mlePanelSizer)
    parentSizer.Add(self.mlePanel)
    w,h = parentSizer.GetMinSize()
    parent.SetVirtualSize((w,h))
    print("parent sizer type = %s  size = %d, %d" %(type(parentSizer).__name__, w, h))
    parent.Layout()
    # call restorePanel to sync data to GUI
    self.restorePanel(currentList, pdef.MaxHashes)
    self.panelCreated = True

  def hidePanel(self):
    """hidePanel - hide the Mle panel"""

    for i in self.mlePanelWidgets:
      i.Hide()


  def showPanel(self):
    """showPanel - show the mle panel"""

    if self.panelCreated:
      for i in self.mlePanelWidgets:
          i.Show()
      parentSizer = self.parent.GetSizer()
      w,h = parentSizer.GetMinSize()
      self.parent.SetVirtualSize((w,h))


  def setElementToDefaults(self):
    """setElementToDefaults - MLE"""

    currentList = self.pdef.getCurrentListObject()
    currentList.ElementDefData[self.myIndex].IncludeInList = False
    currentList.ElementDefData[self.myIndex].HashAlg = 0
    currentList.ElementDefData[self.myIndex].SinitMinVersion = 0
    currentList.ElementDefData[self.myIndex].HashAlg = 0
    currentList.ElementDefData[self.myIndex].Control = 0
    currentList.ElementDefData[self.myIndex].NumbHashes = 0
    currentList.ElementDefData[self.myIndex].CurrentView = 0
    currentList.ElementDefData[self.myIndex].HashFiles = []

  def onOverridePsPolicy(self, event):
    currentList = self.pdef.getCurrentListObject()
    self.setListModified()
    # set/clear bit 0 per MLE Dev Guide PolEltControl def
    if(event.Checked() == True):
      currentList.MleLegacyDefData[self.myIndex].Control = 1
    else:
      currentList.MleLegacyDefData[self.myIndex].Control = 0

  def onMinSinitVersion(self, event):
    value = event.GetString()
    currentList = self.pdef.getCurrentListObject()
    self.setListModified()
    currentList.ElementDefData[self.myIndex].SinitMinVersion = int(value)
    #print("MLE::onMinSinitVersion - SinitMinVersion = %d" % (currentList.ElementDefData[self.myIndex].SinitMinVersion))  # DBGDBG

  # only supports adding files with the add button, not entering the names directly
  def onHashListBoxEdit(self, event):
    """onHashListBoxEdit"""
    #print("in MLE::onHashListBoxEdit")       # DBGDBG

    # tell the user to add button and clear the field
    self.StatusBar.SetStatusText("Please clear the entry, then use the Add button to add hash files to the list")
    #self.hashListBox.Undo()
    #TODO: wxPython: onHashListBoxEdit - Clear() and Undo() both generate an event causing: RuntimeError: maximum recursion depth exceeded

  def onAddButtonClick(self, event):
    """onAddButtonClick - add a hash file to the list"""

    workdir = self.pdef.WorkingDirectory
    filepath, filename = self.selectFile()
    
    if (filename == ''):
      # selectFile() operation has been cancelled.
      return
      
    # validate that the specified hash file is properly formatted
    currentList = self.pdef.getCurrentListObject()
    result = utilities.verifyHashFile(os.path.join(filepath, filename), currentList.ElementDefData[self.myIndex].HashAlg)
    # Note: verifyHashFile() returns [FileValid, FileType], but only FileValid is used here
    if( result[0] == False):
      return

    self.copyFile(filepath, filename)
    self.StatusBar.SetStatusText("Validated file %s." % (filename))

    # update the file count and enable the remove button
    lineCnt = currentList.ElementDefData[self.myIndex].NumbHashes
    #print("onAddButtonClick: was %i lines" % (lineCnt))   # DBGDBG
    lineCnt += 1

    # MaxHashes = 0 means there is no limit on the number of hash files allowed
    if(self.pdef.MaxHashes != 0):
      if(lineCnt > self.pdef.MaxHashes):
        self.StatusBar.SetStatusText("Only %d files can be added." % (self.pdef.MaxHashes))
        return

    #print("onAddButtonClick: now %i lines" % (lineCnt))   # DBGDBG
    self.fileCntEdit.ChangeValue(str(lineCnt))
    self.removeButton.Enable(True)

    hashFileList = currentList.ElementDefData[self.myIndex].HashFiles
    hashFileList.append(filename)
    #print("new hashFileList = %s" % (hashFileList))   # DBGDBG

    # insert the new file into mle.HashFiles
    self.setListModified()
    currentList.ElementDefData[self.myIndex].HashFiles = hashFileList
    currentList.ElementDefData[self.myIndex].NumbHashes = lineCnt
    #print("MleLegacyDefData.NumbHashes=%i, HashFiles = %s" % (currentList.MleLegacyDefData[DEFINES.DEFDATA_INDEX['SHA1']].NumbHashes, currentList.MleLegacyDefData[DEFINES.DEFDATA_INDEX['SHA1']].HashFiles))   # DBGDBG

    # since hashListBox.AppendText() generates an event to onHashListBoxEdit()
    # and since hashListBoxEdit has to be enabled so text can be selected for Remove
    # and direct text entry by the user into hashListBoxEdit is not supported due the complexity of validating it ...
    #
    # hashListBox.ChangeValue() doesn't generate an event but only takes a string, not a hashFileList which is a list ie '[]'
    # so form a single string containing everything in hashFileList and update hashListBox using ChangeValue(hashFileString)
    hashFileString = ''
    index = 0
    for eachString in hashFileList:
      if(index != 0):               # if not the 1st entry, need a LF before the new entry
        hashFileString += "\n"
      hashFileString += eachString
      index += 1
      #print("thisString=%s, hashFileString=%s" % (eachString, hashFileString))

    self.hashListBox.ChangeValue(hashFileString)

  def onRemoveButtonClick(self, event):
    """onRemoveButtonClick - remove the selected entry from the hash file list"""

    # confirm the remove
    dlg = wx.MessageDialog(None, "Confirm removal of selected file?", 'Confirm Remove', wx.YES_NO | wx.ICON_QUESTION)
    response = dlg.ShowModal()
    dlg.Destroy()

    if(response == wx.ID_NO):
      self.StatusBar.SetStatusText( "Remove cancelled" )
      return

    selection = self.hashListBox.GetStringSelection()
    self.StatusBar.SetStatusText("Removed selection %s" % (selection))

    # selection may not be a full line ... See if the selection is contianed in any of hashFileList's entries
    currentList = self.pdef.getCurrentListObject()
    hashFileList = currentList.ElementDefData[self.myIndex].HashFiles
    for entry in hashFileList:
      #print("entry=%s" % (entry))       # DBGDBG
      start = entry.find(selection)
      if(start != -1):          # -1 means not found, else find returns the starting index of selection
        #print("Found: %s at %i" % (selection, start))   # DBGDBG
        # entry was found, but was it a partial selection?
        if(selection not in hashFileList):   # is selection on the GUI in PDEF's hashFileList?
          # partial selection
          #print("Partial selection %s not found in %s." % (selection, hashFileList))    # DBGDBG
          self.StatusBar.SetStatusText("Please select the entire line")
          break
        else:
          # Full selection, so remove that entry
          # decr mle.NumbHashes & update NumberOfFiles widget
          fileCnt = int(self.fileCntEdit.GetValue())
          fileCnt -= 1
          self.fileCntEdit.ChangeValue(str(fileCnt))
          currentList.ElementDefData[self.myIndex].NumbHashes = fileCnt

          hashFileList.remove(selection)          # remove the selection from the PDEF object
          #print("hashFileList=%s" % (hashFileList))        # DBGDBG

          if(fileCnt == 0):                       # disable REMOVE if no more files left
            self.removeButton.Enable(False)
            self.hashListBox.ChangeValue('')
          else:
            # rebuild the content of hashFileEdit from hashFileList and update the screen with ChangeValue
            # to avoid generating an event and to clear the previous LF
            hashFileString = ''
            index = 0
            for eachString in hashFileList:
              if(index != 0):               # if not the 1st entry, need a LF before the new entry
                hashFileString += "\n"
              hashFileString += eachString
              index += 1
              #print("thisString=%s, hashFileString=%s" % (eachString, hashFileString))    # DBGDBG

            self.hashListBox.ChangeValue(hashFileString)

          #print("hashListBox=%s" % (self.hashListBox.GetValue())) # DBGDBG
          self.setListModified()
          currentList.ElementDefData[self.myIndex].HashFiles = hashFileList
          break
    else:
      self.StatusBar.SetStatusText("Selection %s not found. Please select only a single line" % (selection))    # DBGDBG

  def writeMleDef(self, mleDefData, f):
    """writeMleDef - write the Mle Def to the specified file"""

    #print("writeMleDef dump")  # DBGDBG
    pickle.dump(mleDefData, f)       # write out the mleDefData object

  def setPanelToDefaults(self):
    """setPanelToDefaults - restore defaults to mle panel widgets"""

    self.minSinitVersionEdit.ChangeValue("0")
    self.addButton.Enable(True)
    self.removeButton.Enable(False)
    self.fileCntEdit.ChangeValue("0")
    self.hashListBox.ChangeValue("")


  def restorePanel(self, currentList, maxHashes):
    """restorePanel - restore the MLE element panel from the specified PLIST_DEF"""

    print("restorePanel - MLE")   # DBGDBG
    # update Override PS Policy checkbox
    self.overridePsPolicy.SetValue(currentList.ElementDefData[self.myIndex].Control)
    listversion = str(currentList.ListVersionMajor)+'.'+str(currentList.ListVersionMinor)
    if listversion == '2.0':
      self.showV20Gui(True)
    else:
      self.showV20Gui(False)
    self.minSinitVersionEdit.ChangeValue(str(currentList.ElementDefData[self.myIndex].SinitMinVersion))

    # If MaxHashes not 0, Only enable Add if < MaxHashes files
    numbHashes = currentList.ElementDefData[self.myIndex].NumbHashes
    flag = True
    if(maxHashes != 0):
      if(numbHashes >= maxHashes):
        flag = False                              # don't enable add
    self.addButton.Enable(flag)

    # enable remove if >0 hashes
    flag = True
    if(numbHashes > 0):
      flag = False
    self.removeButton.Enable(flag)
    self.fileCntEdit.ChangeValue(str(numbHashes))

    # form a string from hashFileList and update hashListBox
    string = utilities.formStringFromListOfStrings(currentList.ElementDefData[self.myIndex].HashFiles)
    self.hashListBox.ChangeValue(string)

  #def setListModified(self):
  #  """setListModified - if list not modified yet, increment its rev cnt and set it to modified"""
  #
  #  currentList = self.pdef.getCurrentListObject()
  #  #print("Mle setListModified - ListModified was %s" % (currentList.ListModified))  # DBGDBG
  #  if(currentList.ListModified == False):
  #    currentList.RevocationCounter += 1
  #    self.listPanel.revocationCountEdit.ChangeValue(str(currentList.RevocationCounter))   # update the GUI
  #    currentList.ListModified = True




#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# using print() built infunction, disable print statement
from __future__ import print_function

try:
  import os, sys
except ImportError:
  raise ImportError, "import OS failed"

from defines import DEFINES
from struct import pack, unpack
import array
import M2Crypto

from asn1spec import *

# utility routines
class UTILS( object ):

  def checkFileTag(self, file, expectedTag):
    """checkFileTag - Return True if the specified file starts with the specified tag, else False"""

    # expectedTag either:
    #         PCRD+TAG = "PCRD" for .pcr files
    #         HASH_TAG = "HASH" for .hash files
    # read the 1st 4 bytes from the file (i.e. the actualTag) and compare that to the expectedTag
    #
    tagLength = 4
    myFile = open(file, "rb")
    actualTag = myFile.read(tagLength)
    myFile.close()
    #print("UTILS::checkFileTag: file=%s expectedTag=%s actualTag=%s" % (file, expectedTag, actualTag)) # DBGDBG

    if(actualTag == expectedTag):
      return True
    else:
      return False

  def invalidHashFileMsg(self, file, actHashAlg, expHashAlg, hashFileLength):
    """invalidHashFileMsg - display status bar msg if hash file is invalid"""

    print ("Invalid hash format for this element's HashAlg. file %s, Exp Alg=%d, Act Alg=%d, Length=%d" %
              (file, expHashAlg, actHashAlg, hashFileLength))

  # Some functions, such as Wx's ListBox.ChangeValue() takes a string
  # But the data is stored as a list of strins
  # so form a single string containing everything in list
  def formStringFromListOfStrings(self, list):
    """formStringFromListOfStrings - return a single string formed from a list of strings"""

    string = ''
    index = 0
    for eachString in list:
      if(index != 0):               # if not the 1st entry, need a LF before the new entry
        string += "\n"
      string += eachString
      index += 1
      #print("thisString=%s, string=%s" % (eachString, string))

    return string

  def formFileName(self, base, ext):
    """formFileName - return name formed from base.ext"""

    seperator = '.'
    filenameList = [base, ext]
    return(seperator.join(filenameList))


  # return True if file verified and key was extracted, else False
  def verifyKeyFile(self, file, type, currentList):
    """verifyKeyFile - verify the key file and extract the key"""

    # Need currentList.KeySize and currentList.PubKeyQx
    key = self.getKeyFromFile(file, type)
    expectedKeySize = int(currentList.KeySize) / 8

    if type == DEFINES.KEY_FILE_TYPE['PUBLIC_ECDSA']:
      # ECC public key is two value of the key size
      match = len(key) == expectedKeySize *2
      if match:
        # key is flipped for little endian so the latter half is the Qx
        currentList.PubKeyQx = key[expectedKeySize:expectedKeySize*2]
        currentList.PubKeyQy = key[0:expectedKeySize]

    elif type == DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']:
      match = len(key) == expectedKeySize
      if match:
        currentList.PubKeyData = key
    else:
      keylength = len(key)
      match = len(key) == expectedKeySize

    if not match:
      print("Key size mismatch actual=%d expected=%d" %(len(key), expectedKeySize))
    #if(type == KEY_FILE_TYPE_PRIVATE):
    #  expectedHeader = "-----BEGIN RSA PRIVATE KEY-----"
    #  expectedFooter = "-----END RSA PRIVATE KEY-----"
    #  expectedType = "Private"
    #elif(type == KEY_FILE_TYPE_PUBLIC):
    #  expectedHeader = "-----BEGIN PUBLIC KEY-----"
    #  expectedFooter = "-----END PUBLIC KEY-----"
    #  expectedType = "Public"
    #else:
    #  print("invalid key file parameter = %d specified" % (type)) # DBGDBG
    #  return False
    #
    #expHdrLen = len(expectedHeader)
    #try:
    #  f = open(file, "rb")
    #  actualHeader = f.read(expHdrLen)
    #except:
    #  self.StatusBar.SetStatusText("Error opening key file %s"  % (file))
    #  # open failed, can't close
    #  return False
    #
    #if(actualHeader == expectedHeader):
    #  self.StatusBar.SetStatusText("")      # clear any previous error msgs
    #else:
    #  self.StatusBar.SetStatusText("File %s is not a %s key file. Expected header %s" % (file, expectedType, expectedHeader))
    #  f.close()
    #  return False
    #
    ## key file header is OK
    ## read the rest of the file, including line endings
    ## then strip out the line endings and the footer
    ##     Expect Unix style line emdings: 0x0A"  per openssl .pem format
    ##     but also handle Windows style line endings: 0x0D, 0x0A
    ## max file length for keySize=3072 is <3072 since could have different
    ## number of characters on each line and different len endings (0A or 0D0A)
    ## just don't want to read in a [potentially very large] user specified file size
    #maxLength = 3072
    #f.seek (0, os.SEEK_END)
    #actualFileSize = f.tell() - expHdrLen - 1   # size of the rest of the file
    #if(actualFileSize < maxLength):
    #  maxLength = actualFileSize
    #
    ## read the base64 encoded data from the key file into a binary string
    #f.seek (expHdrLen+1, os.SEEK_SET)     # back to just after the header
    #try:
    #  rawData = f.read(maxLength)
    #except:
    #  self.StatusBar.SetStatusText("Error reading key data from file %s" % (file))
    #  f.close()
    #  return False
    #
    ##print("verifyKeyFile: read %d bytes:\n%s" % (len(rawData), rawData)) # DBGDBG
    ##print("verifyKeyFile: read %d" % (len(rawData))) # DBGDBG
    ## Verify the footer
    #if expectedFooter in rawData:
    #  self.StatusBar.SetStatusText("")      # clear any previous error msgs
    #else:
    #  self.StatusBar.SetStatusText("File %s is not a %s key file. Expected footer %s" %
    #                               (file, expectedType, expectedFooter))
    #  f.close()
    #  return False
    #
    ## strip off the footer
    #footerPos = rawData.find('-')    # find 1st '-'
    #rawData = rawData[0:footerPos-1]
    ##print("verifyKeyFile: w/o footer %d bytes:\n%s" % (len(rawData), rawData)) # DBGDBG
    ##print("verifyKeyFile: w/o footer %d bytes" % (len(rawData))) # DBGDBG
    #
    ## Verify that the file matches the current KeySize
    ##
    ## Key File sizes:     1024            2048              3072       for key + footer incl. LR/CR
    ## public file range:  240-260         420-440           590-610
    ## private file range: 850-870         1640-1685         2390-2430
    ##
    #keySize = self.keySizeEdit.GetValue()
    ##print("verifyKeyFile: keySize = %s, type = %s, .pem's length = %d" % (keySize, expectedType, maxLength))  # DBGDBG
    #currentList = self.pdef.getCurrentListObject()
    #misMatch = False
    #if(expectedType == "Public"):
    #  if(keySize == "1024"):
    #    if not ((240 <= maxLength) and (maxLength <= 260)):
    #      print("verifyKeyFile: Public key file, size not 1024!")  # DBGDBG
    #      misMatch = True
    #  elif(keySize == "2048"):
    #    if not ((420 <= maxLength) and (maxLength <= 440)):
    #      print("verifyKeyFile: Public key file, size not 2048!")  # DBGDBG
    #      misMatch = True
    #  elif(keySize == "3072"):
    #    if not ((590 <= maxLength) and (maxLength <= 610)):
    #      print("verifyKeyFile: Public key file, size not 3072!")  # DBGDBG
    #      misMatch = True
    #  if(misMatch == True):
    #    self.StatusBar.SetStatusText("Key size mismatch. File %s is not %s." % (file, keySize))
    #    self.pubKeyFileEdit.SetValue("")
    #    f.close()
    #    return False
    #  else:
    #    keyData = self.decodePublicKeyModulus(file, keySize, expectedHeader, expectedFooter)
    #    currentList.PubKeyData = keyData      # save the public key modulus
    #    #print("verifyKeyFile: Public Key. Length: %d. Data:\n%s" % (len(keyData), currentList.PubKeyData)) # DBGDBG
    #
    #elif(expectedType == "Private"):
    #  if(keySize == "1024"):
    #    if not ((850 <= maxLength) and (maxLength <= 870)):
    #      print("verifyKeyFile: Private key file, size not 1024!")  # DBGDBG
    #      misMatch = True
    #  elif(keySize == "2048"):
    #    if not ((1640 <= maxLength) and (maxLength <= 1685)):
    #      print("verifyKeyFile: Private key file, size not 20481")  # DBGDBG
    #      misMatch = True
    #  elif(keySize == "3072"):
    #    if not ((2390 <= maxLength) and (maxLength <= 2430)):
    #      print("verifyKeyFile: Private key file, size not 3072!")  # DBGDBG
    #      misMatch = True
    #  if(misMatch == True):
    #    self.StatusBar.SetStatusText("Key size mismatch. File %s is not %s." % (file, keySize))
    #    self.pvtKeyFileEdit.SetValue("")
    #    f.close()
    #    return False
    #  # Note: Don't need to save the private key data, pycrypto reads it from the file directly
    #
    #f.close()
    return match


  # Verify that the file is a valid HASH_FILE with a HASH_TAG, hashAlg and hashSize
  #     [with 1 or more SHA1 hashes]  - legacy support from the TPM1.2 LCP1 Spec
  # or a raw data file [i.e. no header] with ONLY 1 20 bit SHA1 or 32 bit SHA256 hash
  #
  # if OK, return a list containing: [True, HashFileModeXXX]  defined in defines.py
  # else return a list containing:   [False, HashFileModeNull]
  #     if file format or length is invalid
  #     or if the length indicates the hash does not correspond to the elements.HashAlg
  #
  #TODO: Bill: verifyHashFile- LCP2 spec  deletes the HASH_FILE struct def for files with multiple hashes??
  #TODO: Bill: verifyHashFile- still supporting files with multiple SHA1 hashes [with the hdr] or raw SHA1 or 256 files
  #
  def verifyHashFile(self, file, hashAlg):
    """verifyHashFile - return a list indicating if the file is valid and its type"""

    # HASH_FILE's are structured as:
    #   - SHA1 or SHA256 hash data only, for raw SHA1 or SHA256 files with 1 hash
    #   - SHA1 hash files with a header [defined below] containing 1 or more SHA1 hashes
    #
    #  Where the header is:
    #     HASH_TAG = 0x48534148 = "HASH"
    #     typedef struct {
    #       UINT32 tag HASH_TAG;          # 4 bytes  "HASH"
    #       UINT8 hashAlg ;               # 1 byte   SHA1_ALG = 4
    #       UINT8 hashSize ;              # 1 byte   SHA1 = 20
    #       UINT8 reserve[10] ;           # 12 bytes
    #       SHA1 - SHA256 hash ;          # 20 bytes
    #     } HASH_FILE;                    # ________
    #     File Size                       # 36 bytes
    # read the 1st 4 bytes from the file (i.e. the actualTag) and compare that to the expectedTag
    #
    function = "verifyHashFile"
    mode = DEFINES.HashFileMode['RawSHA256']    # default

    hashFileLengthHdrSha1   = 36    # header + SHA1 hash data
    hashFileLengthRawSha1   = DEFINES.DIGEST_SIZE['SHA1']    # raw hash files have only the hash data, no header
    hashFileLengthRawSha256 = DEFINES.DIGEST_SIZE['SHA256']
    #hashFileLengthRawSha384 = 48
    #hashFileLengthRawSha512 = 64

    try:
      f = open(file, "rb")
    except:
      print("Error opening hash file %s"  % (file))
      return [False, DEFINES.HashFileMode['HdrNull']]

    try:
      hashAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
      if hashAlgStr == 'SHA1_LEGACY':
          hashAlgStr = 'SHA1'
    except StopIteration:
      print("Unsupported hash algorithm (%d)" %(hashAlg))
      return [False, DEFINES.HashFileMode['HdrNull']]

    #
    # check the file size to determine the type of file being read
    #
    f.seek (0, os.SEEK_END)
    actualSize = f.tell()

    if(actualSize == hashFileLengthHdrSha1):
      mode = DEFINES.HashFileMode['HdrSHA1']
      hashFileLength = hashFileLengthHdrSha1

      # read the data
      #
      f.seek (0, os.SEEK_SET)     # back to the begininng
      data = array.array ("B")   # Load file into data array
      try:
        data.fromfile (f, hashFileLength)
      except:
        print ("Error reading hash from file %s" % (file))
        f.close()
        return [False, DEFINES.HashFileMode['HdrNull']]

      if(hashAlg != DEFINES.TPM_ALG_HASH['SHA1'] and hashAlg != DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
        self.invalidHashFileMsg(file, hashAlg, DEFINES.TPM_ALG_HASH['SHA1'], hashFileLength)
        return [False, DEFINES.HashFileMode['HdrNull']]

      expectedHashTag = "HASH"        # 0x48415348
      expectedSha1HashAlg =  4          # '\x04'
      expectedSha1HashSize = 20         # '\x14'

      actualHashTag  = data[0:4].tostring()          # 0:3 inclusive = 0:4 exclusive of 4
      actualHashAlg  = data[4]
      actualHashSize = data[5]

      if(actualHashTag != expectedHashTag):
        # check if a raw file matching the element's HashAlg length
        print ("File: %s invalid tag = %s, expected %s" % (file, actualHashTag, expectedHashTag))
        return [False, DEFINES.HashFileMode['HdrNull']]

      if(actualHashAlg != expectedSha1HashAlg):
        print ("File: %s invalid hashAlg = 0x%x, expected 0x%x" %
              (file, actualHashAlg, expectedSha1HashAlg))
        return [False, DEFINES.HashFileMode['HdrNull']]

      if(actualHashSize != expectedSha1HashSize):
        print ("File: %s invalid hashSize = 0x%x, expected 0x%x" %
              (file, actualHashSize, expectedSha1HashSize))
        return [False, DEFINES.HashFileMode['HdrNull']]

    elif actualSize == DEFINES.DIGEST_SIZE[hashAlgStr]:
      modeStr = 'Raw' + hashAlgStr
      mode = DEFINES.HashFileMode[modeStr]
      hashFileLength = DEFINES.DIGEST_SIZE[hashAlgStr]
    else:
      self.invalidHashFileMsg(file, hashAlg, DEFINES.TPM_ALG_HASH[hashAlgStr], actualSize)
      print ("File: %s invalid size = %d, expected %d" % (file, actualSize, DEFINES.DIGEST_SIZE[hashAlgStr]))
      return [False, DEFINES.HashFileMode['HdrNull']]

    print("verifyHashFile - HashAlg=%d, Mode=%d, Len=%d" % (hashAlg, mode, hashFileLength)) # DBGDBG

    f.close()
    # handle SHA1 files with headers
    #if(mode == DEFINES.HashFileMode['HdrSHA1']):
    #
    #else:
    #  print("verifyHashFile - Error: invalid mode = %d, aborting." % (mode))
    #  return [False, DEFINES.HashFileMode['HdrNull']]

    return [True, mode]


  #     Determine the type of and validate the pcrFile, except for the numHashes field
  #     This includes checking that the file's type and hashAlg matches the element's expected HashAlg
  #     Return [True, PcrFileModePcrXShaYYY if ok,   see defines.py
  #            [False, PcrFileModeNull]     othewise
  #
  def verifyPcrFile(self, file, elementExpAlg):
    """verifyPcrFile - Validate the pcrFile"""

    # 2 types of PCR dump files are supported: PCRD and PCR2
    # PCR Dump File format
    #     typedef struct {
    #       UINT32 tag PCRD_TAG;          # 4 bytes  "PCRD" = 0x44524350
    #       UINT8 hashAlg ;               # 1 byte   SHA1_ALG = 4 SHA256_ALG = 0x0b
    #       UINT8 hashSize ;              # 1 byte   SHA1 = 20  SHA256 = 32
    #       UINT8 numHashes ;             # 1 byte   number of hashes in the file
    #       UINT8 reserve[9] ;            # 9 bytes
    #       SHA1 pcrs[24] ;               # 20 bytes * numHashes
    #     } HASH_FILE;                    # ________
    #  File Size                       # 16  + (NumHashes * HashSize) bytes
    #  Typically all 24 PCRs included  so size for SHA1 = 16 + 24*20 = 496 = 0x1f0
    #  LCP tool only requires the 1st 8 PCRs, if they are selected via pcr0to7SelectionBitMask
    #  I.e. if the bit mask 0-7 = 1 then that PCR is required
    #

    # - PCR2 PCR dump File format  - from App. C.1.2
    #     typedef struct {
    #       UINT32 tag PCR2_TAG;          # 4 bytes  "PCR2" = 0x32524350
    #       UINT16 hashAlg ;              # 2 bytes  TPM_ALG_SHAXXXXX
    #       UINT8 count ;                 # 1 byte   number of valid digests
    #       UINT8 reserve ;               # 1 byte
    #       UINT16 size ;                 # 2 bytes size of hash
    #     union {
    #       HASH20 pcrs[24] ;             # 20 bytes * 24 of which count are used per the pcrSelection mash
    #       HASH32 pcrs[24] ;             # 32 bytes * 24 of which count are used per the pcrSelection mash
    #     } HASH_FILE;                    # ________
    #  File Size                       # 8 + (count * HashSize) bytes, assuming no holes in the mask
    function = "verifyPcrFile"
    expectedPcrdTag = "PCRD"        # 0x44524350
    expectedPcr2Tag = "PCR2"        # 0x32524350

    try:
      f = open(file, "rb")
    except:
      print ("Error opening PCR data file %s" % (file))
      return [False, DEFINES.PcrFileMode['Null']]

    # Determine if this is a PCRD or PCR2 file
    # Based on the type, ensure the file's HaskAlg matches the element's ExpAlg
    f.seek (0, os.SEEK_SET)     # back to the begininng
    data = array.array ("B")    # Load file into data array
    try:
      data.fromfile (f, DEFINES.PCRFileMinHdrSize)
    except:
      print ("Error reading PCR data from file %s" % (file))
      f.close()
      return [False, DEFINES.PcrFileMode['Null']]

    try:
      elementExpAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if val == elementExpAlg).next()
    except StopIteration:
      print ("Unsupported hash algorithm %d" %(elementExpAlg))

    fileHashTag  = data[0:4].tostring()          # 0:3 inclusive = 0:4 exclusive of 4
    numHashes = data[6]                          # PCRD 'numHashes' same as PCR2 'count'
    fileType = DEFINES.PcrFileMode['Null']
    if(fileHashTag == expectedPcrdTag):          # PCRD file
      fileType = DEFINES.PcrFileMode['PcrdSHA1']
      fileActualHashAlg  = data[4]               # HashAlg is a UINT8
      if(elementExpAlg != DEFINES.TPM_ALG_HASH['SHA1']):
        print ("%s file %s SHA1 hash algorithm does not match the PCONF element's expected algorithm: 0x%x" %
            (expectedPcrdTag, file, elementExpAlg))
        return [False, DEFINES.PcrFileMode['Pcrd']]
      else:
        actAlg = DEFINES.TPM_ALG_HASH['SHA1']    # LCP2 PCRD format only supports SHA1 per C.1.1
        minFileLength   = DEFINES.PCRDFileHdrSize + (8*DEFINES.DIGEST_SIZE['SHA1'])   # min PCRD SHA1   file has 8 hash's

    elif(fileHashTag == expectedPcr2Tag):        # PCR2 file
      # HashAlg in [4:5] is little endian so for current algs high byte=data[5]=0 and low byte=hashAlg
      fileActualHashAlg = unpack('<H', bytearray(data[4:6]))[0]
      #print("verifyPcrFile - elementExpAlg=%d, fileActualHashAlg=%x%x" % (elementExpAlg, fileActualHashAlgHi,fileActualHashAlgLow))  # DBGDBG
      if(elementExpAlg == fileActualHashAlg):
        actAlg = fileActualHashAlg
        fileType = DEFINES.PcrFileMode['Pcr2']
        minFileLength   = DEFINES.PCR2FileHdrSize + (8*DEFINES.DIGEST_SIZE[elementExpAlgStr])   # min PCR2 SHA1   file has 8 hash's
      else:
        print("%s file: %s hashAlg: 0x%x%x does not match the PCONF element's expected algorithm: 0x%02x" %
            (expectedPcr2Tag, file, data[4], data[5], elementExpAlg))
        return [False, DEFINES.PcrFileMode['Null']]

    else:
      print ("File: %s invalid tag = %s, expected %s or %s" % (file, fileHashTag, expectedPcrdTag, expectedPcr2Tag))
      return [False, DEFINES.PcrFileMode['Null']]

    print("verifyPcrFile: %s, tag=%s alg=%i, numHashes=%i" % (file, fileHashTag, actAlg, numHashes))   # DBGDBG

    # check the min file size
    f.seek (0, os.SEEK_END)
    actualFileSize = f.tell ()
    if(actualFileSize < minFileLength):
      print ("File: %s invalid size = %d, expected >= %d" % (file, actualFileSize, minFileLength))
      f.close()
      return [False, DEFINES.PcrFileMode['Null']]

    # file looks ok, except for the numHashes field which is checked by  build.hashPcrInfoFromFile() when it reads the data
    # can't check that yet since don't know which PCRs will be selected
    f.close()
    return [True, fileType]

  #  Verify that the pcr file contains the expected number of pcr hashes
  #  Return True if ok or if no PCR's are selected, else return False
  #
  #  file - is the specified pcr file
  #  fileType - indicates the type of PCR data file per the PcrFileModeXXXXX defines
  #  pcr0to7SelectionBitMask - is a mask indicating the selected PCRs
  #
  def verifyPcrInfoNumHashes(self, file, fileType, hashAlg, pcr0to7SelectionBitMask):
    """verifyPcrInfoNumHashes - verify the pcr file contains enough hashes"""
    function = "verifyPcrInfoNumHashes"

    # check for case when no PCR's were selected
    # nothing to hash in that case
    if(pcr0to7SelectionBitMask == 0):
      print("Warning: No PCR was selected. Nothing to hash: pcr0to7SelectionBitMask=0")
      print ("Note: No PCR was selected. If desired, select PCR's and click 'Apply PCR Selection'")
      return True

    try:
      hashAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if val == hashAlg).next()
    except StopIteration:
      print("verifyPcrInfoNumHashes - invalid fileType=%x hash alg!!!" % (file, fileType, hashAlg))  # Should NEVER get here

    if(fileType == DEFINES.PcrFileMode['Pcrd']):
      minFileLength   = DEFINES.PCRDFileHdrSize + (8*DEFINES.DIGEST_SIZE['SHA1'])   # min PCRD SHA1   file has 8 hash's
    elif(fileType == DEFINES.PcrFileMode['Pcr2']):
      minFileLength   = DEFINES.PCR2FileHdrSize + (8*DEFINES.DIGEST_SIZE[hashAlgStr])   # min PCR2 SHA1   file has 8 hash's
    else:
      print("verifyPcrInfoNumHashes - invalid fileType=%x!!!" % (file, fileType))  # Should NEVER get here

    print("verifyPcrInfoNumHashes %s, type %d, pcr0to7SelectionBitMask=0x%x" % (file, fileType, pcr0to7SelectionBitMask)) # DBGDBG
    try:
      f = open(file, "rb")
    except:
      print ("Error opening PCR data file %s" % (file))
      return False

    f.seek (0, os.SEEK_END)
    actualFileSize = f.tell ()
    f.seek (0, os.SEEK_SET)     # back to the begininng
    data = array.array ("B")    # Load file into data array
    try:
      data.fromfile (f, minFileLength)
    except:
      print ("Error reading PCR data from file %s" % (file))
      f.close()
      return False

    f.close()
    numHashes = data[6]  # same for both PCRD and PCR2 files

    # now that we have NumHashes, check the actual file size
    #print("verifyPcrInfoNumHashes - numHashes=%x" % (numHashes)) # DBGDBG
    if(numHashes == 0):
      print ("File: %s invalid numHashes = 0" % (file))
      return False

    if(fileType == DEFINES.PcrFileMode['Pcrd']):
      expectedFileSize = DEFINES.PCRDFileHdrSize + (numHashes * DEFINES.DIGEST_SIZE['SHA1'])
    elif(fileType == DEFINES.PcrFileMode['Pcr2']):
      expectedFileSize = DEFINES.PCR2FileHdrSize + (numHashes * DEFINES.DIGEST_SIZE[hashAlgStr])
    else:
      print("verifyPcrInfoNumHashes - invalid fileType=%d" % (fileType)) # DBGDBG Should NEVER get here
      return False

    if(actualFileSize < expectedFileSize):
      print ("File: %s invalid File size = 0x%x=%i, expected 0x%x" %
            (file, actualFileSize, actualFileSize, expectedFileSize))
      return False

    # verify that numHashes >= the largest selected hash
    # must have at least that many hashes in the PCR dump file
    #print("verifyPcrInfoNumHashes - verify numHashes are in the file") # DBGDBG
    mask = 0x80
    bit = 8
    while(bit > 0):
      if(mask & pcr0to7SelectionBitMask):
        break                                   # found the largest selected PCR
      else:
        bit -= 1
        mask >>= 1

    if(bit > numHashes):                        # not enough hashes in the PCR dump file, abort
      print ("Too few hashes in PCR dump file: %s, numHashes = 0x%x, but max selected hash = 0x%x" %
            (file, numHashes, bit))
      return False

    return True


  # Get hash data from the specified file and return it using global _GlobalHashData
  # Note: no file format checking is done here as files were validated when added to the list
  # Return False if the hash couldn't be extracted, else return True
  #
  def getHashFromFile(self, file, hashAlg):
    """getHashFromFile - validate the hash file and extract the hash data"""

    global _GlobalHashData

    print("getHashFromFile - file %s, HashAlg=%d" % (file, hashAlg))  # DBGDBG
    result = self.verifyHashFile(file, hashAlg)
    # verifyHashFile() returns a list with [status, fileType]
    fileType = result[1]
    if(result[0] == False):
      return []      # should have detected this when file was selected
    fileType = result[1]
    data = array.array ("B")   # Load file into data array

    try:
      hashAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      print("Unsupported hash algorithm (%d)" %(hashAlg))

    if(fileType == DEFINES.HashFileMode['HdrNull']):
      print("getHashFromFile - Error: invalid mode = %d, aborting." % (result[1]))
      return []
    elif(fileType == DEFINES.HashFileMode['HdrSHA1']):
      hashFileLength = 36
    elif(fileType in DEFINES.HashFileMode.values()):  # All other raw modes
      hashFileLength = DEFINES.DIGEST_SIZE[hashAlgStr]
    else:
      print("getHashFromFile - Error: invalid mode = %d, aborting." % (result[1]))
      return []

    try:
      f = open(file, "rb")
      f.seek (0, os.SEEK_SET)       # beginning of file
      data.fromfile (f, hashFileLength)
    except:
      print("getHashFromFile - Error reading hash from file %s" % (file))
      f.close()
      return []

    f.close()
    #print("getHashFromFile - data = %s, len=%d" % (data, len(data))) # DBGDBG

    # handle all the flavors of hash files
    if(fileType == DEFINES.HashFileMode['HdrSHA1']):
      _GlobalHashData = data[16:36].tolist()        # 20 bytes 16:36 exclusive
      print("getHashFromFile: %s, Hdr tag=%s alg=%i, size=%i, hash=%s, len=%d" %
           (file, data[0:4].tostring(), data[4], data[5], _GlobalHashData, len(_GlobalHashData)))   # DBGDBG
    else:
      _GlobalHashData = data.tolist()
      assert len(data) == DEFINES.DIGEST_SIZE[hashAlgStr], "Error: File size (%d bytes) mismatch with %s size (%d bytes)" %(len(data), hashAlgStr, DEFINES.DIGEST_SIZE[hashAlgStr])
      print("getHashFromFile: %s, raw %s hash=%s, len=%d" % (file, hashAlgStr, _GlobalHashData, len(_GlobalHashData)))   # DBGDBG

    return _GlobalHashData;

  #  Get the specified PCR data (indicated by the thisPcr mask)
  #  Return False if couldn't read the hashes from the PCR file
  #  othewise return True
  #
  #  file - is the specified pcr file
  #  fileType - indicates the type of PCR data file per the PcrFileModeXXXXX defines
  #  thisPcr - the specified PCR data to find  [0 < thisPcr < 8]
  #  hashSize - is the size of the pcr data to extract
  #
  def getPcrInfoFromFile(self, file, fileType, thisPcr, hashSize, statusBar):
    """getPcrInfoFromFile - get the pcr data from the specified file"""

    global _GlobalPcrHash

    if(fileType == DEFINES.PcrFileMode['Pcrd']):
      hdrSize = DEFINES.PCRDFileHdrSize
    elif(fileType == DEFINES.PcrFileMode['Pcr2']):
      hdrSize = DEFINES.PCR2FileHdrSize
    else:
      print("verifyPcrInfoNumHashes - invalid fileType=%d" % (fileType)) # DBGDBG Should NEVER get here
      return False

    #print("getPcrInfoFromFile %s type %d, pcr %d, hashSize=%d" % (file, fileType, thisPcr, hashSize)) # DBGDBG
    pcrData = array.array ("B")
    try:
      f = open(file, "rb")
    except:
      print("getPcrInfoFromFile: open of file %s failed" % (file)) # should never get here, file's been opened for verif. earlier
      return False

    # read TPM1 or TPM2 PCR tag to determine the file format to use?
    tag = f.read(4)

    # read the selected pcr's hash data
    if tag == 'PCR2':
      # PCR2 file format
      pos = hdrSize + (thisPcr * (hashSize + 2))  # There's a 2 byte size field precedes each PCR value.
      f.seek (pos, os.SEEK_SET)
      # Read the size of the PCR value and compare to the expected hash size
      size = f.read(2)
      pcrSize = unpack("<H", size)[0]
      if pcrSize == hashSize:
        pos += 2
      else:
        print ("Error reading hash from file %s at position %d with size %d" % (file, pos, hashSize))
        f.close()
        return False
    elif tag == 'PCRD':
      # PCR1 file format
      pos = hdrSize + (thisPcr * hashSize)
    else:
      print ("Error invalid PCR tag format for file %s" % (file))

    print("getPcrInfoFromFile - Read hash %d @pos%d, size %d" % (thisPcr, pos, hashSize)) # DBGDBG
    try:
      # calculate the start of the selected pcr's data
      f.seek (pos, os.SEEK_SET)     # back to the next hash where pos depends on if PCRD or PCR2 file
      #print("getPcrInfoFromFile - Seek done") # DBGDBG
      pcrData.fromfile(f, hashSize)
    except:
      print ("Error reading hash from file %s at position %d with size %d" % (file, pos, hashSize))
      f.close()
      return False

    f.close()
    _GlobalPcrHash = pcrData[0:hashSize].tolist()    # extract the data
    #print("getPcrInfoFromFile: %s, _GlobalPcrHash=%s" % (file, _GlobalPcrHash)) # DBGDBG
    return _GlobalPcrHash


  #     Open pcrFile
  #     Copy selected PCR measurements (pcrSelect) to temp buffer
  #     calculate composite hash of temp buffer
  #     return the composite hash in _GlobalPcrHash
  #     return False if couldn't read the hashes from the PCR file
  #     othewise return True [including if pcr0to7SelectionBitMask = 0 so nothing to hash]
  #
  def hashPcrInfoFromFile(self, file, pcr0to7SelectionBitMask, numSelectedPcrs):
    """hashPcrInfoFromFile - hash the pcr data from the specified file and return the hash"""
    # TODO: This function is use by PCONF legacy only?

    # PCR Dump File format
    #     typedef struct {
    #       UINT32 tag PCRD_TAG;          # 4 bytes  "PCRD" = 0x44524340
    #       UINT8 hashAlg ;               # 1 byte   SHA1_ALG = 4 SHA256_ALG = 0x0b
    #       UINT8 hashSize ;              # 1 byte   SHA1 = 20  SHA256 = 32
    #       UINT8 numHashes ;             # 1 byte   number of hashes in the file
    #       UINT8 reserve[9] ;            # 9 bytes
    #       SHA1 pcrs[24] ;               # 20 bytes * numHashes
    #     } HASH_FILE;                    # ________
    #  File Size                       # 16  + (NumHashes * HashSize) bytes
    #  Typically all 24 PCRs included  so size for SHA1 = 16 + 24*20 = 496 = 0x1f0
    #  LCP tool only requires the 1st 8 PCRs, if they are selected via pcr0to7SelectionBitMask
    #  I.e. if the bit mask 0-7 = 1 then that PCR is required
    #

    global _GlobalPcrHash
    #_GlobalPcrHash   = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]  # DBGDBG
    _GlobalPcrHash   = b'0000000000000000000000000000000000000000'

    # check for case when no PCR's were selected
    # nothing to hash in that case
    if(pcr0to7SelectionBitMask == 0):
      print("Warning: No PCR was selected. Nothing to hash: pcr0to7SelectionBitMask=0")
      print ("Note: No PCR was selected. If desired, select PCR's and click Update")
      return True

    try:
      f = open(file, "rb")
    except:
      print ("Error opening PCR data file %s" % (file))
      return False

    # Get the file size of PCR dump
    f.seek (0, os.SEEK_END)
    actualFileSize = f.tell ()

    f.seek (0, os.SEEK_SET)     # back to the begininng
    header = array.array("B")   # Read PCR header to determine type
    data = array.array ("B")    # Load file into data array

    # Read the header portion.  16-bytes should cover current PCRD and PCR2 file format.
    try:
      header.fromfile (f, 16)   # PCRD header size is 16 bytes.  PCR2 header size is 8 bytes.
    except:
      print ("Error reading PCR data from file %s" % (file))
      f.close()
      return False

    # Read the number of valid hashes
    numHashes = header[6]

    # If there are no hash in the dump file, the PCR dump is not valid
    if(numHashes == 0):
      print ("File: %s invalid numHashes = 0" % (file))
      return False

    # now that we have NumHashes, check the actual file size
    # Check the tag in the PCR file
    tag = header[0:4].tostring()
    if tag == 'PCR2':
      expectedPcrHashSize = unpack("<H", header[8:10])[0]                       # every measurement in the PCR dump file should be the same size
      pcrStartPos = 8                                                           # position of the first PCR measurement in the file
      minFileLength = pcrStartPos + (8 * (2 + expectedPcrHashSize))             # PCR2 file structure has 8 bytes followed by size and value for each measurement. min file has 8 hash's
      expectedFileSize = pcrStartPos + (numHashes * (2 + expectedPcrHashSize))  # include the 2-byte size field for each PCR value
    elif tag == 'PCRD':
      expectedPcrHashSize = 0x14                                                # for TPM1 only SHA1 is supported which has length of 20 bytes.
      pcrStartPos = 16                                                          # position of the first PCR measurement in the file
      minFileLength = pcrStartPos + (8 * expectedPcrHashSize)                   # PCRD file structure has 16 bytes followed by value of each measurements. min file has 8 hash's
      expectedFileSize = pcrStartPos + (numHashes * expectedPcrHashSize)

    # Check PCR dump for valid file size
    if(actualFileSize < expectedFileSize):
      print ("File: %s invalid FileSize = 0x%x=%i, expected 0x%x" %
            (file, actualFileSize, actualFileSize, expectedFileSize))
      return False

    # read min pcr values
    try:
      f.seek (0, os.SEEK_SET)
      data.fromfile (f, minFileLength)
    except:
      print ("Error reading PCR data from file %s" % (file))
      f.close()
      return False
    f.close()

    # verify that numHashes >= the largest selected hash
    # must have at least that many hashes in the PCR dump file
    mask = 0x80
    bit = 8
    while(bit > 0):
      if(mask & pcr0to7SelectionBitMask):
        break                                   # found the largest selected PCR
      else:
        bit -= 1
        mask >>= 1

    if(bit > numHashes):                        # not enough hashes in the PCR dump file, abort
      print ("Too few hashes in PCR dump file: %s, numHashes = 0x%x, but max selected hash = 0x%x" %
            (file, numHashes, bit))
      return False

    # file looks ok, get the data
    temp = array.array ("B")      # temp array
    hashLength = expectedPcrHashSize
    numSelectedPcrs = 0

    mask = 0x01
    bit = 0
    pos = 0
    while(bit < 8):
      if(mask & pcr0to7SelectionBitMask):
        if tag == 'PCR2':
          pos = pcrStartPos + 2 + bit * (2 + hashLength)
        elif tag == 'PCRD':
          pos = pcrStartPos + bit * hashLength

        temp += data[pos:pos+hashLength]
        numSelectedPcrs += 1
        print("hashPcrInfoFromFile - Read hash %d, mask 0x%x from file %s, select=0x%x, numSelectedPcrs=%d, len(temp)=%d" %
              (bit, mask, file, pcr0to7SelectionBitMask, numSelectedPcrs, len(temp))) # DBGDBG

      mask <<= 1
      bit += 1
      pos += hashLength

    #print("hashPcrInfoFromFile TypeOf: tempList=%s, tempList[0]=%s, _GlobalPcrHash=%s, _GlobalPcrHash[0]=%s" %
    #        (type(tempList), type(tempList[0]), type(_GlobalPcrHash), type(_GlobalPcrHash[0])))  # DBGDBG
    #pcrHash = hashlib.sha1()
    hashAlg = header[4]   # TODO: is hashAlg determined by GUI or from PCR file.
    hashAlgStr = None
    try:
      hashAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      print ("Error unsupported hash algorithm (%d)" %(hashAlg))

    # Set hash algorithm
    pcrHash = None
    if 'SM3' in hashAlgStr:
      pcrHash = sm3()
    else:
      pcrHash = M2Crypto.EVP.MessageDigest(hashAlgStr.lower())

    if tag == 'PCR2':
      pcrHash.update(temp)
    elif tag == 'PCRD':
      #pcrHash = hashlib.sha1()
      pcrHash = M2Crypto.EVP.MessageDigest('sha1')
      # The PCR composite hash consists of:  TPM_PCR_COMPOSITE structure
      #     UINT16  sizeOfSelect              # BigEndian = 00 03
      #     UINT8   pcr_select[3]
      #     UINT32  valueSize                 # BigEndian = 20 * NumberOfSelectedHashes
      #     UINT8   pcrValue[]                # all the selected PCR hashes
      data = pack("<BBBBB", 0, 3, pcr0to7SelectionBitMask, 0, 0)  # pack sezeOfSelect and pcr_select[3]
      pcrHash.update(data)
      valueSize = numSelectedPcrs * DEFINES.DIGEST_SIZE['SHA1']
      data = pack(">L", valueSize)            # Note '>' for BigEndian packing of valueSize
      pcrHash.update(data)
      # pack pcrValue[]
      pcrHash.update(temp)

    # hash.digest() Returns the digest of the strings passed to the update() method so far.
    # This is a string of digest_size bytes [which may contain non-ASCII characters, including null bytes]
    # Note: cannot pass this string thru struct.pack() which takes ints
    _GlobalPcrHash = pcrHash.digest()

    #print("hashPcrInfoFromFile: %s, Generated hash: Length=%d HexData=%s " %
    #      (file, hashLength, pcrHash.hexdigest()))   # DBGDBG

    return _GlobalPcrHash


  def getKeyFromFile(self, file, type):
    # Read the key file from the
    with open(file, 'rb') as kf:
       mb = M2Crypto.BIO.MemoryBuffer(kf.read())

    kf.close()
    pem = mb.getvalue()
    pem_lines = pem.split('\n')
    key = None
    # Find line index of BEGIN and END header/footer
    der = ''
    foundbegin = False
    for line in pem_lines:
      if ('END' in line) and ('PUBLIC' in line or 'PRIVATE' in line):
        break
      if foundbegin:
        der += line.strip()
      if ('BEGIN' in line) and ('PUBLIC' in line or 'PRIVATE' in line):
        foundbegin = True

    try:  # in case ASN1 can't decode the ASN1 notation..
      if type == DEFINES.KEY_FILE_TYPE['PRIVATE_RSASSA']:
        asn, substrate = der_decoder.decode(der.decode('base64'), asn1Spec=RSAPrivateKey())
        rsapvt = asn.getComponentByName('privateExponent')   # returns univ.Integer()
        octet = univ.OctetString(hexValue=format(int(rsapvt), '0x'))  # convert to Octet
        key = octet.asOctets()
      elif type == DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']:
        # This decodes DER encoded ASN1 public key
        asn, substrate = der_decoder.decode(der.decode('base64'), asn1Spec=SubjectPublicKeyInfo())
        bits = asn.getComponentByName('subjectPublicKey')

        # second level decode for RSAPublicKey()
        bits_string = ''.join(map(str, bits))
        octet = univ.OctetString(binValue=bits_string)

        rsaasn, substrate = der_decoder.decode(octet.asOctets(), asn1Spec=RSAPublicKey())
        rsapub = rsaasn.getComponentByName('modulus')     # returns univ.Integer()
        octet = univ.OctetString(hexValue=format(int(rsapub), '0x'))  # convert to Octet
        key = octet.asOctets()

      elif type == DEFINES.KEY_FILE_TYPE['PRIVATE_ECDSA']:
        asn, substrate = der_decoder.decode(der.decode('base64'), asn1Spec=ECPrivateKey())
        ecpvt = asn.getComponentByName('privateKey')      # returns univ.OctetString()
        key = ecpvt.asOctets()

      elif type == DEFINES.KEY_FILE_TYPE['PUBLIC_ECDSA']:
        # This decodes DER encoded ASN1 public key
        asn, substrate = der_decoder.decode(der.decode('base64'), asn1Spec=SubjectPublicKeyInfo())
        bits = asn.getComponentByName('subjectPublicKey') # returns univ.BitString()

        # DSAPublicKey is Integer() so no decoding is needed, but need to remove the prefix 0x04.
        bits_string = ''.join(map(str, bits[8:]))  # the first byte specifies the compress alg?
        octet = univ.OctetString(binValue=bits_string)
        key = octet.asOctets()

    except Exception as ex:
      print ("Exception: unable to decode pem file")
      print (ex)

    if key != None:
      keyLE = key[::-1]  # little endian
    else:
      keyLE = ''

    return keyLE


  # the last function in the file doesn't show up in the scope list in Understand for some reason!
  def stub(self):
    pass



#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

#  TXT Policy Generator Tool - tools invoked from tools menu
#     Hash image
#     Insert policy into image

# using print() built infunction, disable print statement
from __future__ import print_function

try:
  import wx
except ImportWxError:
  raise ImportWxError, "Please download the appropriate version of wxPython from www.wxpython.org"

try:
  import os
except ImportOsError:
  raise ImportOsError, "import OS failed"

import array
import M2Crypto
import struct

# Pycrypto module from: https://www.dlitz.net/software/pycrypto/ is required
#try:
#  from Crypto.Hash import SHA               # SHA1
#  from Crypto.Hash import SHA256            # SHA256
#except ImportError:
#  raise ImportError, "PyCrypto import failed. Please install PyCrypto ..."

from defines import DEFINES
from util import UTILS
utilities = UTILS()


class TOOLS(object):

  # hash a bios image 'file' [i.e. base.bin] from 'startOffset' for 'offsetSize' bytes,
  # and write the hash to base.hash, as a LCP2 raw hash file, and base.dat as an 'ascii coded binary'
  # I.e. if the image's hash starts with:  0xB0, 0x61, 0x52, ...
  # then  the    .hash file  starts with:  B0 61 52 ...   (i.e. the raw binary hash)
  # and   the    .dat  file  starts with:  42 30 20 36 31 20 35 32 20
  # (i.e. each hex digit is represented by its ascii equivalent with an ascii space after each hex pair

  # Parameters
  #   biosFileName - name of the bios file to be hashed
  #   startOffset  - offset into the file to start hashing from, aka base
  #   offsetSize   - size to hash from startOffset
  #   hashAlg      - hash algorithm to use, TPM_ALG_XXXX where XXXX is SHA1 or SHA256

  # Return True on success or
  #        False on any error
  #
  def hashImage(self, biosFileName, startOffsetStr, offsetSizeStr, hashAlgStr):
    """hashImage - generate the hash of the specified file from start address for size bytes"""
    function = 'hashImage'

    biosPath = os.path.abspath(os.path.normpath(biosFileName))
    startOffset = int(startOffsetStr, 16)
    offsetSize  = int(offsetSizeStr, 16)
    hashAlg     = int(hashAlgStr, 16)

    MAX_BIOS_SIZE = 0x1000000;
    print("%s - biosFile=%s, startOffset=0x%x, offsetSize=0x%x, hashAlg=%d" %
          (function, biosPath, startOffset, offsetSize, hashAlg))  # DBGDBG

    # use biosFile's base name for the .tmp and .hash files
    #   biosFile - file to be hashed, ie if biosFileName = biosXYZ.bin then base = biosXYZ
    #   tempFile - base.tmp - temp file containing the data to hash
    #     i.e. biosFile's data from startOffset to startOffset+offsetSize
    #   hashFile - base.hash - output file containg the hash of tempFile

    # verify that biosFile can be opened
    try:
      biosFile = open(biosPath, 'rb')
    except:
      print("Unable to open specified file: %s" % (biosPath))
      return False

    # strip off the .ext leaving the base
    if(biosFileName.endswith('.bin') == True):
      base = os.path.basename(biosFileName).split('.bin')
    else:
      print("Expected .bin extention for Bios File %s" % (biosPath))
      return False

    #TODO:  will all bios files have .bin extentions??????????????????????????????????????
    #print("%s - base=%s" % (function, base[0]))  # DBGDBG

    # verify that .tmp, .dat and .hash files can be created in the current dir, else exit
    tmpFileName  = utilities.formFileName(base[0], "tmp")
    hashFileName = utilities.formFileName(base[0], "hash")
    datFileName  = utilities.formFileName(base[0], "dat")
    #print("%s - tmpFileName=%s, hashFileName=%s" % (function, tmpFileName, hashFileName)) # DBGDBG
    try:
      tmpFile = open(tmpFileName, 'wb')
    except:
      print("Unable to create file: %s" % (tmpFileName))
      biosFile.close()
      return False

    try:
      hashFile = open(hashFileName, 'wb')
    except:
      print("Unable to create file: %s" % (hashFileName))
      tmpFile.close()
      biosFile.close()
      return False

    try:
      datFile = open(datFileName, 'wb')
    except:
      print("Unable to create file: %s" % (datFileName))
      tmpFile.close()
      biosFile.close()
      datFile.close()
      return False

    # Determine biosFile's size
    biosFile.seek (0, os.SEEK_END)
    biosFileSize = biosFile.tell()

    # if startOffset > MAX_BIOS_SIZE, then adjust startOffset to be relative to 4Gb
    if (startOffset > MAX_BIOS_SIZE):
      print("The specified StartOffset(0x%x) is > than the max bios size(%x) Assuming StartOffset is relative to 4Gb" %
        (startOffset, MAX_BIOS_SIZE))
      beginingOfBiosImage = 0x100000000 - biosFileSize;
      startOffset = startOffset - beginingOfBiosImage;
      print("%s - startOffset now 0x%x, MAX_BIOS_SIZE=0x%x" % (startOffset, MAX_BIOS_SIZE))

    # If startOffset + offsetSize > biosFile's size, adjust offsetSize
    if(startOffset + offsetSize > biosFileSize):
      print("%s - The specified StartOffset(0x%x) + offsetSize(0x%x) is larger than biosFileSize(0x%x)" %
        (function, startOffset, offsetSize, biosFileSize))
      offsetSize = biosFileSize - startOffset;
      print("StartOffset+OffsetSize > BiosFileSize. Resetting OffsetSize to 0x%x" % (offsetSize))

    #print("%s - biosFileSize=0x%x, startOffset+offsetSize=0x%x, MAX_BIOS_SIZE=0x%x" % (function, biosFileSize, startOffset+offsetSize, MAX_BIOS_SIZE)) # DBGDBG
    # buffer the part of the bios that will be hashed in base.tmp
    #   open the bios file,
    #   set the read pointer to: startOffset
    #   read offsetSize bytes into tempFile
    data = array.array ("B")
    biosFile.seek(startOffset, os.SEEK_SET)
    #print("%s - pos=0x%x" % (function, biosFile.tell())) # DBGDBG
    data.fromfile(biosFile, offsetSize)

    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    # reverse lookup of the hash algorithm name(key) for the given HashAlg value
    hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == hashAlg)).next()

    if (hashAlgName != None):
      # This handles all SHA hash algorithms.
      hash = M2Crypto.EVP.MessageDigest(hashAlgName.lower())
    else:
      self.StatusBar.SetStatusText("Hash Algorithm %d is not supported" % (hashAlg))
      return False

    hash.update(data)
    hashdigest = hash.digest()  # cannot call this M2Crypto function twice.

    print(data, end='', file=tmpFile )        # for testing that the right data was hashed
    print(hashdigest, end='', file=hashFile )
    hexHash = hashdigest.encode('hex')
    print("Image's hash: %s" % (hexHash))
    print("Generated hash files: %s.hash and %s.dat" % (base[0],base[0]))

    # generate the .dat from hexHash: convert each hex byte to a ascii digit followed by a space
    i=0
    while(i < len(hexHash)):
      print(hexHash[i].upper(),   end='', file=datFile )    # 1st hexdigit as upper case ascii
      print(hexHash[i+1].upper(), end='', file=datFile )    # 2nd hexdigit as upper case ascii
      print(' ',          end='', file=datFile )            # space after each pair
      i += 2
    print('\n', end='', file=datFile )                      # final LF to match WinLCP

    biosFile.close()
    tmpFile.close()
    hashFile.close()
    datFile.close()
    return True



#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# using print() built infunction, disable print statement
from __future__ import print_function

#
# wxPython is not part of the standard Python distribution and has to be downloaded and installed separately.
# Tell the user that wxPython is required but has not been found
#
try:
  import wx
except ImportError:
  raise ImportError, "Please download the appropriate version of wxPython from www.wxpython.org"


import os
import shutil

from defines import DEFINES

#
# TXT Policy Generator Tool
#
class ElementGui( object ):

  CONST_TITLE = "Choose file"
  CONST_WILDCARD = "All Files (*.*)  | *.*"

  def __init__( self ):
    pass


  # isElementType() compares the name argument to the class name and stored hash algorithm name
  # to find a matching element.
  # The name argument should have a format of <element name>-<hash alg name>
  def isElementType(self, name):
    result = False
    type = name.split('-')
    if len(type) != 2:
      result = False
    else:
      elementType, hashAlg = type
      if hashAlg == 'LEGACY':
        if elementType in self.__class__.__name__:
          result = True
      else:
        if elementType in self.__class__.__name__ and self.myHashAlg == DEFINES.TPM_ALG_HASH[hashAlg]:
          result = True
    return result


  def getHashAlgName(self):
    if 'Legacy' in self.__class__.__name__:
      name = 'SHA1-LEGACY'
    else:
      try:
        name = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.myHashAlg)).next()
      except StopIteration:
        name = None
    return name


  # getName() is intended to get the class name from the classes defined in pdef.py
  #
  #def getName(self):
  #  if 'Legacy' in self.__class__.__name__:
  #    hashname = 'LEGACY'
  #    name = self.__class__.__name__.split('Legacy')[0] + '-' + hashname
  #  else:
  #    try:
  #      hashname = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.myHashAlg)).next()
  #      name = self.__class__.__name__.split('_DEF')[0] + '-' + hashname
  #    except StopIteration:
  #      name = None
  #  return name

  def selectFile(self):
    """onAddButtonClick - This code is common to all elements to select a file
       and copy file into current working directory if the file doesn't exist """

    filepath = ''
    filename = ''
    workdir = self.pdef.WorkingDirectory

    dlg = wx.FileDialog(self.parent, self.CONST_TITLE, workdir, "", self.CONST_WILDCARD, wx.FD_OPEN)

    if dlg.ShowModal() == wx.ID_CANCEL :
      self.StatusBar.SetStatusText( "Add cancelled" )
    else:
      filename = dlg.GetFilename()
      filepath = dlg.GetDirectory()

    dlg.Destroy()
    # return null string indicate file select cancelled.
    return filepath, filename


  def copyFile(self, filepath, filename):
    workdir = self.pdef.WorkingDirectory

    if (filepath != workdir):
      if (os.path.exists(os.path.join(workdir, filename))) :
        dlg = wx.MessageDialog(self.parent, filename+" already exists in working directory\nOverwrite file in working directory?", "Confirm Copy", wx.OK|wx.CANCEL|wx.ICON_QUESTION)

        if (dlg.ShowModal() == wx.ID_OK):
          shutil.copyfile(os.path.join(filepath, filename), os.path.join(workdir, filename))
          self.StatusBar.SetStatusText( "File copied" )
        else:
          self.StatusBar.SetStatusText( "File copy aborted" )

        dlg.Destroy()
      else:
        shutil.copyfile(os.path.join(filepath, filename), os.path.join(workdir, filename))

        
  def setListModified(self):
    """setListModified - if list not modified yet, increment its rev cnt and set it to modified"""

    currentList = self.pdef.getCurrentListObject()

    #print("PCONF setListModified - ListModified was %s" % (currentList.ListModified))  # DBGDBG
    if(currentList.ListModified == False):
      currentList.RevocationCounter += 1
      self.listPanel.revocationCountEdit.ChangeValue(str(currentList.RevocationCounter))   # update the GUI
      currentList.ListModified = True
    self.pdef.Modified = True


  def showV20Gui(self, enable):
    if enable:
      self.overridePsPolicy.Show()
    else:
      self.overridePsPolicy.Hide()


  def enableDisableOverridePsPolicy(self, value):
    """enableDisableOverridePsPolicy widget"""

    self.overridePsPolicy.Enable(value)


#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# using print() built infunction, disable print statement
from __future__ import print_function

import string

#
# wxPython is not part of the standard Python distribution and has to be downloaded and installed separately.
# Tell the user that wxPython is required but has not been found
#
try:
  import wx
except ImportError:
  raise ImportError, "Please download the appropriate version of wxPython from www.wxpython.org"

try:
  import os
  import shutil
except ImportError:
  raise ImportError, "import OS failed"

from defines import DEFINES
from pdef import PLIST_DEF, MLE_DEF, PCONF_DEF, SBIOS_DEF
from LcpPolicy import LCP_POLICY2, LCP_SIGNATURE2
from util import UTILS
utilities = UTILS()

from sbios import SBIOS
from mle import MLE
from pconf import PCONF
from stm import STM
from sbiosLegacy import SBIOSLegacy
from pconfLegacy import PCONFLegacy
from mleLegacy import MLELegacy
from pdef import *


try:
    import cPickle as pickle
except ImportError:
    import pickle         # fall back on Python version


import array
import base64
from struct import *


# TXT Policy Generator Tool
# List Class - Policy Definition File Lists
#
class LIST( object ):
  """__init__() - List class constructor"""
  def __init__( self ):
    """LIST __init__ constructor"""
    self.listPanelWidgets = []
    self.visibleElement = DEFINES.ELEMENT_NAME_NONE
    self.includedElements = []
    #print("in LIST __init__() listPanelWidgets=%s" % (self.listPanelWidgets))   # DBGDBG

  #
  # create the List Panel
  #     LIST X
  #       Version 2.0   Number of 0 Select  NONE      DELETE ELEMENT
  #                     Elements    Element
  #       Revocation 23 [] Signed                     ADD SBIOS
  #       Count            Key File my.key  BROWSE    ADD MLE
  #       Allowed    23    Key Size 2048  RSA PKCS1.5 ADD PCONF
  #      RESET   [] Sync                  Algorithm
  #
  def createListPanel(self, wx, parent, listNumber, pdef, statusBar):
    """createListPanel - create the List Panel"""

    #print("createListPanel - list %i" % (listNumber))
    self.pdef = pdef
    self.StatusBar = statusBar
    self.parent = parent
    parentSizer = parent.GetSizer()

    # create the List Panel sizers
    self.listPanel = wx.Panel(parent, -1)
    self.listPanelSizer = wx.GridBagSizer(hgap=5, vgap=5)
    self.listPanel.SetSizer(self.listPanelSizer)
    #listHorizSizer = wx.BoxSize4r(wx.HORIZONTAL)

    #listLabelText = "LIST " + str(listNumber)
    listLabelText = "LIST"
    self.listLabel = wx.StaticText(self.listPanel, -1, listLabelText)
    font = wx.Font( 18, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    self.listLabel.SetFont( font )
    self.listPanelSizer.Add( self.listLabel, pos=(0, 0))
    self.listPanelWidgets = [self.listLabel]

    self.listLabelNum = wx.TextCtrl( self.listPanel, value="X", size=(30, -1))
    self.listLabelNum.Enable( False )              # LIST num cannot be modified by user
    self.listLabelNum.SetFont( font )
    #self.listLabelNum.SetBackgroundColour(self.listPanel.GetBackgroundColour())            # no change to color??   Fore or Background N/C
    self.listLabelNum.SetFocus()
    self.listLabelNum.Refresh()   # no help
    self.listLabelNum.ChangeValue(str(listNumber))
    self.listPanelSizer.Add( self.listLabelNum,  pos=(0,1))
    self.listPanelWidgets.append(self.listLabelNum)

    # LIST Version text box - PDEF.PolList[n].ListVersion
    # Change list version number to corresponds to policy version number
    #policyversion = str(pdef.PolVersionMajor) + '.' + str(pdef.PolVersionMinor)
    #listversion = DEFINES.SUPPORTED_LCP_VERSION[policyversion]
    #majorstring, minorstring = listversion.split('.')
    #self.pdef.PolListInfo[str(listNumber-1)].ListVersionMajor = int(majorstring)
    #self.pdef.PolListInfo[str(listNumber-1)].ListVersionMinor = int(minorstring)
    self.versionLabel = wx.StaticText(self.listPanel, label="Version: ", size = (90, -1), style = wx.ALIGN_RIGHT)
    self.listPanelSizer.Add( self.versionLabel, pos=(1,0))
    self.listPanelWidgets.append(self.versionLabel)
    version = str(self.pdef.PolListInfo[str(listNumber-1)].ListVersionMajor)+"."+str(self.pdef.PolListInfo[str(listNumber-1)].ListVersionMinor)
    self.versionEdit  = wx.TextCtrl( self.listPanel, value=version, size=(30, -1))
    self.versionEdit.Enable( False )              # Version cannot be modified
    self.listPanelSizer.Add( self.versionEdit,  pos=(1,1))
    self.listPanelWidgets.append(self.versionEdit)

    # Number Elements text box - number of elements that are marked VALID
    self.numberElementsLabel = wx.StaticText(self.listPanel, label="Number of Elements ")
    self.listPanelSizer.Add( self.numberElementsLabel, pos=(2,2))
    self.listPanelWidgets.append(self.numberElementsLabel)
    self.numberElementsEdit  = wx.TextCtrl( self.listPanel, value="0", size=(30, -1))
    self.numberElementsEdit.Disable()
    self.listPanelSizer.Add( self.numberElementsEdit,  pos=(2,3))
    self.listPanelWidgets.append(self.numberElementsEdit)

    # Select Element Control - PDEF.PolList[n].CurrentElementView
    #     Allows user to select which elements to view
    #     Disabled when NumberOfElements = 0
    #     SBIOS is only valid if PS rules.
    #     MLE and PCONF are valid for both.
    self.selectElementList = [DEFINES.ELEMENT_NAME_NONE]

    self.selectElementLabel = wx.StaticText(self.listPanel, label="\nView Element")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.NORMAL)
    self.selectElementLabel.SetFont( font )
    self.listPanelSizer.Add(self.selectElementLabel, pos=(3,2))
    self.listPanelWidgets.append(self.selectElementLabel)

    self.selectElementEdit = wx.ComboBox( self.listPanel, size=(100, -1), value=DEFINES.ELEMENT_NAME_NONE, choices=self.selectElementList, style=wx.CB_READONLY )
    self.selectElementEdit.Enable( False )
    self.selectElementEdit.SetSelection(0)
    self.listPanelSizer.Add(self.selectElementEdit, pos=(4,2))
    self.selectElementEdit.Bind(wx.EVT_TEXT, self.onSelectElementEdit)
    self.listPanelWidgets.append(self.selectElementEdit)

    # ADD ELEMENT
    #     adds the selected element
    #     Disabled when all possible elements exist
    #     sets corresponding IncludeXXX to TRUE
    #
    self.addElementButton = wx.Button( self.listPanel, -1, label="Add    Element")
    self.listPanelSizer.Add( self.addElementButton, pos=(2,0))
    self.listPanelWidgets.append(self.addElementButton)
    self.addElementButton.Bind(wx.EVT_BUTTON, self.onAddElementClick)

    self.addElementChoicesPs = []
    self.addElementChoicesPs += DEFINES.ELEMENT
    self.addElementChoicesPo = []
    self.addElementChoicesPo += DEFINES.ELEMENT_PO_RULES

    # DELETE ELEMENT combo
    #     Deletes the selected element
    #     Disabled when NumberOfElements = 0
    #     sets corresponding IncludeXXX to FALSE
    #
    self.deleteElementButton = wx.Button( self.listPanel, -1, label="Delete Element")
    self.deleteElementButton.Enable( False )
    self.listPanelSizer.Add( self.deleteElementButton, pos=(3,0))
    self.listPanelWidgets.append(self.deleteElementButton)
    self.deleteElementButton.Bind(wx.EVT_BUTTON, self.onDeleteElementButtonClick)

    # Revocation Count text box - PDEF.PolList[n]ListVersion.RevocationCounter
    #     updated automatically each time the user performs a BUID, but only if the list is signed
    self.revocationCountLabel = wx.StaticText(self.listPanel, label="Revocation\nCount: ")
    self.listPanelSizer.Add( self.revocationCountLabel, pos=(2,6))
    self.listPanelWidgets.append(self.revocationCountLabel)
    self.revocationCountEdit  = wx.TextCtrl( self.listPanel, value="0", size=(30, -1))
    self.revocationCountEdit.Enable(False)                  # disable editing since updated automatically on BUILD's
    self.listPanelSizer.Add( self.revocationCountEdit,  pos=(2,7))
    self.listPanelWidgets.append(self.revocationCountEdit)

    # Allowed text box - PDEF.PolList[n]ListVersion.RevokeCount
    #     permits user to enter a value between 0 and RevocationCounter
    #     Copied to PDEF.RevocationCounters[n] when policy is built
    self.allowedLabel = wx.StaticText(self.listPanel, label="Allowed")
    self.listPanelSizer.Add( self.allowedLabel, pos=(3,6))
    self.listPanelWidgets.append(self.allowedLabel)
    self.allowedEdit  = wx.TextCtrl( self.listPanel, value="0", size=(30, -1))
    self.allowedEdit.Enable(False)                       # disabled since Sync checkbox defaults to True
    self.listPanelSizer.Add( self.allowedEdit,  pos=(3,7))
    self.listPanelWidgets.append(self.allowedEdit)
    self.allowedEdit.Bind(wx.EVT_TEXT, self.onAllowedEdit)

    # Reset button
    #     Resets the Revocation Count to 0
    #
    self.resetButton = wx.Button( self.listPanel, -1, label="Reset", style=wx.BU_EXACTFIT)
    self.resetButton.Enable(False)
    self.listPanelSizer.Add( self.resetButton, pos=(4,6))
    self.listPanelWidgets.append(self.resetButton)
    self.resetButton.Bind(wx.EVT_BUTTON, self.onResetButtonClick)

    # Sync checkbox - PDEF.PolListInfo[n].SyncRevCount
    #      if checked, force Allowed to equal Revocation Count and Allowed box is disabled
    #
    self.Sync = wx.CheckBox(self.listPanel, label="Sync")
    self.Sync.SetValue(True)
    self.Sync.Bind(wx.EVT_CHECKBOX, self.onSync)
    self.listPanelSizer.Add(self.Sync,  pos=(4,7), span=(1,2), flag=wx.BOTTOM, border=5)
    self.listPanelWidgets.append(self.Sync)

    # Private Key File text box - PDEF.PolList[n].PvtKeyFile
    #     allows user to specify the file that contains the signing key
    self.pvtKeyFileSel = wx.ComboBox( self.listPanel, size=(80, -1), value="Private Key", choices=["Private Key", "Signature"], style=wx.CB_READONLY )
    self.pvtKeyFileSel.Bind(wx.EVT_TEXT, self.onPvtKeyFileSel)
    self.listPanelSizer.Add( self.pvtKeyFileSel, pos=(2,11))
    self.listPanelWidgets.append(self.pvtKeyFileSel)

    self.pvtKeyFileEdit  = wx.TextCtrl( self.listPanel, value="", size=(110, -1))
    self.pvtKeyFileEdit.Bind(wx.EVT_TEXT, self.onPvtKeyFileEdit)
    self.listPanelSizer.Add( self.pvtKeyFileEdit,  pos=(2,12))
    self.listPanelWidgets.append(self.pvtKeyFileEdit)

    # BROWSE button
    #     bring up a File Browse dialogue
    #
    self.pvtBrowseButton = wx.Button( self.listPanel, -1, label="Browse", style=wx.BU_EXACTFIT)
    self.listPanelSizer.Add( self.pvtBrowseButton, pos=(2,13))
    self.listPanelWidgets.append(self.pvtBrowseButton)
    self.pvtBrowseButton.Bind(wx.EVT_BUTTON, self.onPvtBrowseButtonClick, self.pvtBrowseButton)

    # Key File text box - PDEF.PolList[n].PubKeyFile
    #     allows user to specify the file that contains the signing key
    self.pubKeyFileLabel = wx.StaticText(self.listPanel, label="Public Key")
    #font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.NORMAL)
    #self.pubKeyFileLabel.SetFont( font )
    self.listPanelSizer.Add( self.pubKeyFileLabel, pos=(3,11))
    self.listPanelWidgets.append(self.pubKeyFileLabel)

    self.pubKeyFileEdit  = wx.TextCtrl( self.listPanel, value="", size=(110, -1))
    self.pubKeyFileEdit.Bind(wx.EVT_TEXT, self.onPubKeyFileEdit)
    self.listPanelSizer.Add( self.pubKeyFileEdit,  pos=(3,12))
    self.listPanelWidgets.append(self.pubKeyFileEdit)

    # BROWSE button
    #     bring up a File Browse dialogue
    #
    self.pubBrowseButton = wx.Button( self.listPanel, -1, label="Browse", style=wx.BU_EXACTFIT)
    self.listPanelSizer.Add( self.pubBrowseButton, pos=(3,13))
    self.listPanelWidgets.append(self.pubBrowseButton)
    self.pubBrowseButton.Bind(wx.EVT_BUTTON, self.onPubBrowseButtonClick)

    # Key Size Control - PDEF.PolList[n].KeySize
    #     Allows user to select one of the supported key sizes: 1024, 2048, 3072
    #
    self.keySizeList = DEFINES.SIGNATURE_KEY_SIZE["RSA PKCS1.5/SHA256"]
    self.keySizeLabel = wx.StaticText(self.listPanel, label="Key Size")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.NORMAL)
    self.keySizeLabel.SetFont( font )
    self.listPanelSizer.Add(self.keySizeLabel, pos=(2,10))
    self.listPanelWidgets.append(self.keySizeLabel)

    self.keySizeEdit = wx.ComboBox( self.listPanel, size=(110, -1), value="2048", choices=self.keySizeList, style=wx.CB_READONLY )
    self.keySizeEdit.Bind(wx.EVT_TEXT, self.onKeySizeEdit)
    self.listPanelSizer.Add(self.keySizeEdit, pos=(3,10))
    self.listPanelWidgets.append(self.keySizeEdit)

    # AlgorithmControl - PDEF.PolList[n].SigAlgorithm
    #     Allows user to select supported algorithms, but only RSA PKCS1.5
    #
    self.algorithmList = DEFINES.SIGNATURE_ALGORITHMS
    self.algorithmLabel = wx.StaticText(self.listPanel, label="Signing\nAlgorithm ")
    font = wx.Font( 10, wx.FONTFAMILY_DEFAULT, wx.NORMAL, wx.BOLD)
    self.algorithmLabel.SetFont( font )
    self.listPanelSizer.Add(self.algorithmLabel, pos=(1,11))
    self.listPanelWidgets.append(self.algorithmLabel)

    self.algorithmEdit = wx.ComboBox( self.listPanel, size=(165, -1), value="None", choices=self.algorithmList, style=wx.CB_READONLY )
    self.algorithmEdit.Bind(wx.EVT_TEXT, self.onAlgorithmEdit)
    self.listPanelSizer.Add(self.algorithmEdit, pos=(1,12), span=(1,2))
    self.listPanelWidgets.append(self.algorithmEdit)
    self.enableDisableSigned(False)                     # default to not signed

    #listHorizSizer.Add(listGridSizer,  0, wx.ALL, 5)
    #self.listPanelSizer.Add(listHorizSizer, 0, wx.ALL, 5)
    #parent.Add(listHorizSizer)
    parentSizer.Add(self.listPanel)
    w,h = parentSizer.GetMinSize()
    parent.SetVirtualSize((w,h))
    print("parent sizer type = %s  size = %d, %d" %(type(parentSizer).__name__, w, h))
    #parent.Fit()
    parent.Layout()
    #parent.SetSizerAndFit(self.listPanelSizer)

  def hideListPanel(self):
    """hideListPanel - remove [actually just hide] the list panel and any element panels from the policy panel"""

    for i in self.listPanelWidgets:
        i.Hide()

    self.hideAllPanels()     # hide all the panels

    #TODO: wxPython: hideListPanel - how to resize window back to the original Policy Panel size when delting a list?

  def showListPanel(self):
    """showListPanel - re-show the list and element panels"""

    #print("in showListPanel listPanelWidgets=%s" % (self.listPanelWidgets))   # DBGDBG
    for i in self.listPanelWidgets:
        i.Show()

    for element in self.includedElements:
      element.showPanel()

    parentSizer = self.parent.GetSizer()
    w,h = parentSizer.GetMinSize()
    self.parent.SetVirtualSize((w,h))


  def setListPanelToDefaults(self, wx, parent, listNumber, pdef):
    """setListPanelToDefaults - widgets were all ready created, restore defaults"""
    #self.StatusBar.SetStatusText("Created LIST %i with default values" % (listNumber))
    print("Created LIST %i with default values" % (listNumber))

    self.listLabelNum.ChangeValue(str(listNumber))
    policyversion = str(pdef.PolVersionMajor)+'.'+str(pdef.PolVersionMinor)
    listversion = DEFINES.SUPPORTED_LCP_VERSION[policyversion]
    self.versionEdit.ChangeValue(listversion)
    self.numberElementsEdit.ChangeValue("0")
    self.selectElementEdit.Clear()
    self.selectElementEdit.Append('None')
    self.selectElementEdit.SetSelection(0)
    self.addElementChoicesPs = []
    self.addElementChoicesPs += DEFINES.ELEMENT
    self.addElementChoicesPo = []
    self.addElementChoicesPo += DEFINES.ELEMENT_PO_RULES
    self.revocationCountLabel.Enable(True)
    self.revocationCountEdit.ChangeValue("0")
    self.allowedLabel.Enable(True)
    self.allowedEdit.ChangeValue("0")
    self.resetButton.Enable(True)
    self.Sync.SetValue(True)
    self.Sync.Enable(True)
    self.algorithmEdit.SetValue("None")
    self.pubKeyFileEdit.ChangeValue("")
    self.pubKeyFileEdit.Enable(False)
    self.pvtKeyFileEdit.ChangeValue("")
    self.pvtKeyFileEdit.Enable(False)
    self.pubBrowseButton.Enable(False)
    self.pvtBrowseButton.Enable(False)
    self.keySizeEdit.SetValue("2048")
    self.keySizeEdit.Enable(False)
    self.algorithmEdit.Enable(True)

    for i in self.listPanelWidgets:
        i.Show()

    for elements in self.includedElements:
      elements.hidePanel()

    # delete element GUI panels in self.includedElements[] and elements in pdef.CurrentList.ElementDefData[]
    self.includedElements = []
    currentListObject = pdef.getCurrentListObject()
    currentListObject.ElementDefData = []


  def restoreListPanel(self, currentList):
    """restoreListPanel - restore the list panel's widgets"""
    func = 'restoreListPanel'

    self.hideAllPanels()
    self.listLabelNum.ChangeValue(str(self.pdef.CurrentListView))
    cnt = 0
    self.rebuildSelectElementChoices()
    self.addElementChoicesPs = []
    self.addElementChoicesPs += DEFINES.ELEMENT           # copy elements not just reference
    self.addElementChoicesPo = []
    self.addElementChoicesPo += DEFINES.ELEMENT_PO_RULES  # copy elements not just reference

    cnt = len(currentList.ElementDefData)
    #self.numberElementsEdit.ChangeValue(str(cnt))

    if(cnt > 0):
      flag = True
      self.selectElementEdit.SetValue(currentList.CurrentElementView)
      #self.selectElementEdit.SetSelection(cnt)
    else:
      flag = False
      currentList.CurrentElementView  = DEFINES.ELEMENT_NAME_NONE
      #self.visibleElement = DEFINES.ELEMENT_NAME_NONE
      self.selectElementEdit.SetValue(DEFINES.ELEMENT_NAME_NONE)
      self.selectElementEdit.SetSelection(0)

    # show element panel for selected element
    selection = self.selectElementEdit.GetSelection()
    if selection > 0:
      self.includedElements[selection-1].showPanel()
      self.selectElementEdit.Enable(flag)
      self.deleteElementButton.Enable(flag)

    #print("restoreListPanel - Sync=%s SigAlgorithm=%x" % (currentList.SyncRevCount, currentList.SigAlgorithm))  # DBGDBG
    self.Sync.SetValue(currentList.SyncRevCount)
    self.allowedEdit.ChangeValue(str(currentList.RevokeCounter))
    self.revocationCountEdit.ChangeValue(str(currentList.RevocationCounter))
    self.keySizeEdit.SetValue(str(currentList.KeySize))
    #if (currentList.SigAlgorithm == 1):  # For TPM 1.2

    signAlgName = ""
    try:
      signAlgName = (key for key,val in DEFINES.TPM_ALG_SIGN.items() if currentList.SigAlgorithm == val).next()
    except StopIteration:
      print("WARNING - Invalid signature algorithm (%d)" %(currentList.SigAlgorithm))

    hashAlgName = ""
    try:
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if currentList.sigAlgorithmHash == val).next()
    except StopIteration:
      print("WARNING - Invalid hash algorithm (%d)" %(currentList.sigAlgorithmHash))

    selectionString = "None"
    if 'RSA' in signAlgName:
      selectionString = (name for name in DEFINES.SIGNATURE_ALGORITHMS if 'RSA' in name and hashAlgName in name).next()
      enable = True
    elif 'ECDSA' in signAlgName:
      selectionString = (name for name in DEFINES.SIGNATURE_ALGORITHMS if signAlgName in name and currentList.KeySize in name).next()
      enable = True
    #elif 'SM2' in signAlgName:
    #  selectionString = (name for name in DEFINES.SIGNATURE_ALGORITHMS if signAlgName in name).next()
    #  enable = True
    else:
      enable = False

    self.enableDisableSigned(enable, selectionString)
    self.algorithmEdit.SetValue(selectionString)
    self.pvtKeyFileSel.SetSelection(currentList.PvtKeyFileIsSignature)
    self.pubKeyFileEdit.ChangeValue(currentList.PubKeyFile)
    self.pvtKeyFileEdit.ChangeValue(currentList.PvtKeyFile)
    #currentList.ListModified = False # don't want to change the state of this variable

  def setListPanelToCurrentListView(self):
    """setListPanelToCurrentListView - update the list panel to the specified list"""

    currentList = self.pdef.getCurrentListObject()
    print("setListPanelToCurrentListView %i AllowedCounter=%i" % (self.pdef.CurrentListView, currentList.RevokeCounter)) # DBGDBG

    # Hide all panels before rebuilding the list for the new panel
    self.hideAllPanels()
    # Clear GUI panels for each element
    self.includedElements = []

    # Re-build the GUI panel for each element from the PDEF persistent object in pdef.py file
    index = 0
    for defdata in currentList.ElementDefData:
      # Get Name and Hash type of each and recreate GUI element with it.

      # Create a new element of selected spec and populate into currentListObject.MleDefData
      elementType, hashAlg = defdata.Name.split('-')

      if hashAlg == 'LEGACY':
        if elementType == 'SBIOS':
          element = SBIOSLegacy()
        elif elementType == 'MLE':
          element = MLELegacy()
        elif elementType == 'PCONF':
          element = PCONFLegacy()
        else:
          print ("ERROR: invalid element")
      else:
        if elementType == 'SBIOS':
          element = SBIOS(DEFINES.TPM_ALG_HASH[hashAlg])      # GUI panel
        elif elementType == 'STM':
          element = STM(DEFINES.TPM_ALG_HASH[hashAlg])
        elif elementType == 'MLE':
          element = MLE(DEFINES.TPM_ALG_HASH[hashAlg])
        elif elementType == 'PCONF':
          element = PCONF(DEFINES.TPM_ALG_HASH[hashAlg])
        else:
          print ("ERROR: invalid element")

      element.createOrShowPanel(wx, self, self.parent, self.pdef, self.StatusBar)
      element.myIndex = index
      self.includedElements.append(element)
      element.restorePanel(currentList, self.pdef.MaxHashes)
      element.hidePanel()
      index += 1

    # restore any element panels that exist in the current PLIST_DEF

    # Also update this list's element panel, if one exists
    # 1st hide all the  elements and enable if they don't exist and rules allow that
    #self.hideAllPanels()    # Hide all the element panels

    # then disable add buttons for elements that exist
    # and count the number of elements in this list
    numberOfElements = len(currentList.ElementDefData)

    # rebuild selectElementEdit's choices list for the elements existing in this view
    self.rebuildSelectElementChoices()

    # then show a panel

    # update the NumberOfElements and enable selectElementsEdit if > 0
    self.numberElementsEdit.ChangeValue(str(numberOfElements))
    if(numberOfElements > 0):
        self.selectElementEdit.Enable( True )
        self.deleteElementButton.Enable( True )

    self.restoreListPanel(currentList)


  # only called if PolicyType = LIST if pdef.NumLists > 0
  # for each signed list,
  #   If Sync is checked,
  #     then update the current list's AllowedCounter from its RevocationCounter
  #   update pdef.DataRevocationCounters[list#] from the current list's Allowedounter
  #   if the current List is displayed in the list panel, update the list panel widgets
  #
  # NOTE: the current list's RevocationCounter is incr on its 1st modification
  #
  def onBuildButtonClick(self, pdef):
    """ onBuildButtonClick - perform a build"""
    #print("list::onBuildButtonClick")   # DBGDBG

    listNum = 0
    while( listNum < pdef.NumLists):
      thisList = pdef.PolListInfo[str(listNum)]
      #if(thisList.SigAlgorithm == DEFINES.LCP_POLSALG_RSA_PKCS_15):  # For TPM1.2
      if(thisList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']):
        if(thisList.SyncRevCount == True):
          thisList.RevokeCounter = thisList.RevocationCounter

        pdef.DataRevocationCounters[listNum] = thisList.RevokeCounter

        # is this list being shown in the list panel?
        if(pdef.CurrentListView == listNum + 1):
          self.allowedEdit.ChangeValue(str(thisList.RevokeCounter))
          self.revocationCountEdit.ChangeValue(str(thisList.RevocationCounter))

        if(thisList.RevocationCounter > 0):       # can't reset til cnt > 0
            self.resetButton.Enable(True)

        print("list.onBuild - list %d, RevCnt=%d, RevokeCnt=%d, pdef.DataRevocationCounters[%d]=%d" %
          (listNum+1, thisList.RevocationCounter, thisList.RevokeCounter, listNum+1, pdef.DataRevocationCounters[listNum]))    # DBGDBG

      listNum += 1

    #print("list.onBuild - pdef.DataRevocationCounters=%s" % (pdef.DataRevocationCounters)) # DBGDBG

  def onAllowedEdit(self, event):
    """ onAllowedEdit - handle Allowed value change"""

    string = event.GetString()
    self.StatusBar.SetStatusText("in onAllowedEdit, value=%s" % (string))

    currentList = self.pdef.getCurrentListObject()
    maxValue = currentList.RevocationCounter
    #print("onAllowedEdit - MaxValue=%i" % (maxValue))  # DBGDBG

    try:
      value = int(string)
    except:
      self.StatusBar.SetStatusText(  "%s is invalid, Please enter only digits between 0 and %i" % (string, maxValue))
    else:
      #print("onAllowedEdit - Value=%i" % (value))  # DBGDBG
      if(value > maxValue):
        self.allowedEdit.ChangeValue(str(currentList.RevokeCounter))
        self.StatusBar.SetStatusText( "%i is too large, the max value for Allowed is %i" % ( value, maxValue ))
      else:
        currentList.RevokeCounter = value
        print("onAllowedEdit - setting RevokeCounter=%i MaxValue=%i" % (value, maxValue))
        self.setListModified()

  def onSelectElementEdit(self, event):
    """onSelectElementEdit - view the selected element"""
    func = 'onSelectElementEdit'

    #selection = self.selectElementEdit.GetSelection()
    selection = event.GetEventObject().GetSelection()
    selectionString = event.GetString()
    #self.setListModified()   # just looking at different element shouldn't be considered modifying it.
    #Element = DEFINES.ELEMENT
    #Object  = self.getObject()
    currentList = self.pdef.getCurrentListObject()
    #IncludeElement = utilities.getIncludeElement(currentList)

    # Using exhaustive find method to match selected element by it's element name and hash alg
    #for element in self.includedElements:
    #  found = element.isElementType(selectionString)
    #  if found:
    #    self.StatusBar.SetStatusText( "Displaying the %s element" % (selectionString))
    #    self.hideAllPanels()
    #    element.showPanel()
    #    self.visibleElement = selectionString
    #    currentList.CurrentElementView = selectionString
    #    break

    if(selectionString == DEFINES.ELEMENT_NAME_NONE):
      self.StatusBar.SetStatusText( "No element selected")
      self.visibleElement = selectionString
      currentList.CurrentElementView = selectionString
      self.hideAllPanels()
    else:
      self.StatusBar.SetStatusText( "Displaying the %s element" % (selectionString))
      self.hideAllPanels()
      self.includedElements[selection-1].showPanel()
      self.visibleElement = selectionString
      currentList.CurrentElementView = selectionString


    #TODO: wxPython: onSelectElementEdit - how to resize panel here?  Or set min size after each panel is created?
    print("%s: selection=%d, visibleElement=%s" % (func, self.selectElementEdit.GetSelection(), self.visibleElement)) # DBGDBG)

  def onDeleteElementButtonClick(self, event):
    """ onDeleteElementButtonClick - delete the current element"""
    func = 'onDeleteElementButtonClick'
    self.StatusBar.SetStatusText( "Deleted the current element")

    #exit if None is selected, else confirm the deletion and continue
    if(self.visibleElement == DEFINES.ELEMENT_NAME_NONE):
      self.StatusBar.SetStatusText( "No element selected, Please select the element to be deleted")
      return

    # confirm delete
    dlg = wx.MessageDialog(None, "Deleted Elements cannot be recovered. Continue?", 'Confirm Element Deletion', wx.YES_NO | wx.ICON_QUESTION)
    response = dlg.ShowModal()
    dlg.Destroy()

    if(response == wx.ID_NO):
      self.StatusBar.SetStatusText( "Element Deletion cancelled" )
      return
    else:
      self.StatusBar.SetStatusText( "Element deleted")

    # ComboBox selection value is one index higher than self.includedElement[] index and PLIST_DEF.ElementDefData[] index
    # because of the None element to select no view.
    selection = self.selectElementEdit.GetSelection()             # get the index
    selectionString = self.selectElementEdit.GetValue()           # get selected element and populate into Add Element menu
    self.selectElementEdit.Delete(selection)
    print("%s - Selection=%d VisibleElement=%s - Removed" % (func, selection, self.visibleElement)) # DBGDBG

    currentElementsCnt = int(self.numberElementsEdit.GetValue())
    if(selection > 0):
      nextselection = selection - 1
      # fixup selection for out of order case where selection 1 was deleted, but are > 1 elements left
      # ie should only select 0 if 0 elements left  ie None is the only choice left
      # Scenario:  create > 1 elements, select 1, delete 1. 1 still selected
      if((nextselection == 0) and (currentElementsCnt > 1)):
        nextselection = 1
      self.selectElementEdit.SetSelection(nextselection)
    else:
      print("%s: selectElementEdit.GetSelection() is invalid, Aborting!! % (func)")
      return

    # set IncludeXXX to False
    # hide the current element, show another element if one exists
    self.setListModified()
    #self.dumpIncludeXXX( currentListObject, 'before delete' )      # show current value of all the  XXXDefData[i].IncludeXXXX  # DBGDBG

    # Add deleted element back to Add Element menu.
    self.updateAddElementChoicesAfterDelete(selectionString)

    # update myIndex of the elements behind this element
    for count in range(selection, len(self.includedElements)):
      self.includedElements[count].myIndex -= 1

    # remove item from pdef list
    currentListObject = self.pdef.getCurrentListObject()
    currentListObject.ElementDefData.pop(selection-1)

    # update GUI to show other element
    if selection > 0:
      self.includedElements[selection-1].hidePanel()
      # remove this element's GUI
      self.includedElements.pop(selection-1)

    # nextselection is indexing into the updated self.includedElements[]
    if nextselection != 0:
      self.includedElements[nextselection-1].showPanel()

    # Set other GUI fields
    currentElementsCnt -= 1
    self.numberElementsEdit.ChangeValue(str(currentElementsCnt))
    if(currentElementsCnt > 0):
      self.selectElementEdit.Enable( True )
      self.addElementButton.Enable(True)
      self.visibleElement = self.selectElementEdit.GetValue()
      currentListObject.CurrentElementView = self.selectElementEdit.GetValue()
      print("%s - Selection=%d VisibleElement=%s" % (func, selection, self.visibleElement)) # DBGDBG
    elif(currentElementsCnt == 0):
      self.selectElementEdit.Enable( False )
      self.deleteElementButton.Enable( False )
      self.visibleElement = DEFINES.ELEMENT_NAME_NONE
      currentListObject.CurrentElementView = DEFINES.ELEMENT_NAME_NONE

    #self.dumpIncludeXXX( currentListObject, 'after delete' )      # show current value of all the  XXXDefData[i].IncludeXXXX  # DBGDBG

  # Object - array of all the mle, sbios, pconf & stm 1 and 256 objects
  # Element - array of all the ELEMENT_NAME_XXXX_SHAYYY strings
  # IncludeElement - array of all the XXXXDefData[DEFINES.DEFDATA_INDEX_SHANNN.IncludeXXXX
  #         updated to indicate the just deleted element's value is now False
  #
  def showNextElement(self, Object, Element, IncludeElement, currentListObject):
    """showNextElement - if another element's panel exists, show it"""
    func = 'showNextElement'

    i=0
    while(i < len(Element)):
      #print("%s - checking %s" % (func, Element[i]))  # DBGDBG
      if(IncludeElement[i] == True):
        print("%s - showing %s" % (func, Element[i]))  # DBGDBG
        Object[i].showPanel()
        self.visibleElement = Element[i]
        currentListObject.CurrentElementView = Element[i]
        self.selectElementEdit.SetValue(Element[i])
        break

      i += 1      # next element


  def onResetButtonClick(self, event):
    """ onResetButtonClick - reset the revocation count"""
    self.StatusBar.SetStatusText( "You clicked the Reset Button!")

    if(self.revocationCountEdit.GetValue() != "0"):
      dlg = wx.MessageDialog(None, "Revocation Count reset cannot be undone. Continue?", 'Confirm Revocation Count reset', wx.YES_NO | wx.ICON_QUESTION)
      response = dlg.ShowModal()
      dlg.Destroy()

      if(response == wx.ID_NO):
        self.StatusBar.SetStatusText( "Revocation Count reset cancelled" )
      else:
        self.StatusBar.SetStatusText( "Reset the Revocation Count " )
        self.revocationCountEdit.ChangeValue("0")
        currentList = self.pdef.getCurrentListObject()
        currentList.RevocationCounter = 0
        #self.setListModified()         # don't set list modified here as that re-increments the revocation ctr
        self.resetButton.Enable(False)

  def onSync(self, event):
    """ onSync - force Allowed to equal Revocation Count and disable Allowed"""

    self.StatusBar.SetStatusText("You checked Sync")
    currentList = self.pdef.getCurrentListObject()
    if(event.IsChecked()):
      self.allowedEdit.ChangeValue(self.revocationCountEdit.GetValue())
      currentList.SyncRevCount = True
      currentList.RevokeCounter = currentList.RevocationCounter
      self.setListModified()
      print("onSync - PolList[%i].RevCnt=%i, Allowed=%i" %
        (self.pdef.CurrentListView, currentList.RevokeCounter, currentList.RevocationCounter))    # DBGDBG
    else:
      currentList.SyncRevCount = False

    # when checked disable editing the allowed box
    self.allowedEdit.Enable(not event.IsChecked())

  def enableDisableSigned(self, value, signAlgName='None'):
    """enableDiableSigned - perform common actions when Signed or Algorithm changed"""

    # Note: Sync checkbox and Algorithm pulldown are redundant
    #   Sync unchecked  == LCP_POLSALG_NONE
    #   Sync checked    == LCP_POLSALG_RSA_PKCS_15

    #If Unsigned [ie algorithm=None], then revocation count, allowed, RESET, Sync, Key File & Key Size are all disabled,
    # else enabled

    currentList = self.pdef.getCurrentListObject()
    keysizeList = DEFINES.SIGNATURE_KEY_SIZE[signAlgName]

    if(value == False):
      #value = False
      self.allowedEdit.Enable(value)
      #currentList.SigAlgorithm = DEFINES.TPM_ALG_SIGN['NULL']       # 0=Not signed
      #self.algorithmEdit.SetValue(DEFINES.ELEMENT_NAME_NONE)
    else:
      #value = True
      self.allowedEdit.Enable(not value)

    self.setListModified()
    self.resetButton.Enable(value)
    self.Sync.Enable(value)
    self.pubKeyFileEdit.Enable(value)
    self.pvtKeyFileEdit.Enable(value)
    self.pubBrowseButton.Enable(value)
    self.pvtBrowseButton.Enable(value)
    self.keySizeEdit.SetItems(keysizeList)
    self.keySizeEdit.SetValue(str(currentList.KeySize))
    self.keySizeEdit.Enable(value)
    self.revocationCountLabel.Enable(value)
    self.allowedLabel.Enable(value)
    self.pvtKeyFileSel.Enable(value)
    self.pubKeyFileLabel.Enable(value)
    self.keySizeLabel.Enable(value)

  def onPubKeyFileEdit(self, event):
    """ onpubKeyFileEdit - update the key file"""

    currentList = self.pdef.getCurrentListObject()
    currentList.PubKeyFile = event.GetString()
    self.setListModified()

    # once the entire file name is entered, verify it, else clear it
    if(currentList.PubKeyFile.endswith(".pem")):
      if currentList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']:
        type = DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']
      elif currentList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['ECDSA']:
        type = DEFINES.KEY_FILE_TYPE['PUBLIC_ECDSA']

      if(utilities.verifyKeyFile(currentList.PubKeyFile, type, currentList) == False):
        self.pubKeyFileEdit.ChangeValue("")

  def onPvtKeyFileSel(self, event):
    """ onPvtKeyFileEdit - update the key file"""
    currentList = self.pdef.getCurrentListObject()
    selectionString = event.GetString()
    if selectionString == "Private Key":
      currentList.PvtKeyFileIsSignature = False
    else:
      currentList.PvtKeyFileIsSignature = True


  def onPvtKeyFileEdit(self, event):
    """ onPvtKeyFileEdit - update the key file"""

    currentList = self.pdef.getCurrentListObject()
    currentList.PvtKeyFile = event.GetString()
    self.setListModified()

    # once the entire file name is entered, verify it, else clear it
    if((currentList.PvtKeyFile.endswith(".pem")) or (currentList.PvtKeyFile.endswith(".key"))):
      if(utilities.verifyKeyFile(currentList.PvtKeyFile, KEY_FILE_TYPE_PRIVATE, currentList) == False):
        self.pvtKeyFileEdit.ChangeValue("")

  def onKeySizeEdit(self, event):
    """ onKeySizeEdit - update the key size"""

    currentList = self.pdef.getCurrentListObject()
    currentList.KeySize = event.GetString()
    self.setListModified()
    self.StatusBar.SetStatusText("")   # clear any previous error messages

    #if key file was all rdy specified, verify its size is correct
    file = self.pubKeyFileEdit.GetValue()
    if(file != ""):
      if(utilities.verifyKeyFile(file, KEY_FILE_TYPE_PUBLIC, currentList) == False):
        self.pubKeyFileEdit.ChangeValue("")
        currentList.PubKeyFile = ""

    file = self.pvtKeyFileEdit.GetValue()
    if(file != ""):
      if(utilities.verifyKeyFile(file, KEY_FILE_TYPE_PRIVATE, currentList) == False):
        self.pvtKeyFileEdit.ChangeValue("")
        currentList.PvtKeyFile = ""

  def onPvtBrowseButtonClick(self, event):
    """ onPvtBrowseButtonClick - browse to an existing PDEF file"""

    #self.StatusBar.SetStatusText("You clicked the Browse button")
    #TODO: Bill: handle .pems that have both the public and private keys?
    #TODO: Bill: handle .pems with password protected private keys?

    currentList = self.pdef.getCurrentListObject()
    #accept .key as well as pem files for private keys?
    dirname = ''   #  current working directory
    workdir = self.pdef.WorkingDirectory

    if currentList.PvtKeyFileIsSignature:
      title = "Choose the Signature file"
      wildcard = "All Files (*.*)    | *.*" \
                 "Key file (*.sig) | *.sig|"
    else:
      title = "Choose the Private Key file"
      wildcard = "Key file (*.pem) | *.pem|" \
                 "Key file (*.key) | *.key|" \
                 "All Files (*.*)    | *.*"
    dlg = wx.FileDialog(self.parent, title, workdir, "", wildcard, wx.FD_OPEN)

    if dlg.ShowModal() == wx.ID_OK:
      filename = dlg.GetFilename()
      dirname  = dlg.GetDirectory()

      self.pvtKeyFileEdit.ChangeValue(filename)
      currentList.PvtKeyFile = filename
      self.setListModified()

      if currentList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']:
        filetype = DEFINES.KEY_FILE_TYPE['PRIVATE_RSASSA']
      else:
        filetype = DEFINES.KEY_FILE_TYPE['PRIVATE_ECDSA']

      if currentList.PvtKeyFileIsSignature:
        #copy signature file to the current working directory
        pass
      else:
      # Verify Private Key file
        if(utilities.verifyKeyFile(os.path.join(dirname, currentList.PvtKeyFile), filetype, currentList) == False):
          self.pvtKeyFileEdit.ChangeValue("")
          currentList.PvtKeyFile = ""
        elif (dirname != workdir):
          if (os.path.exists(os.path.join(workdir, filename))) :
            confdlg = wx.MessageDialog(self.parent, filename+" already exists in working directory\nOverwrite file in working directory?", "Confirm Copy", wx.OK|wx.CANCEL|wx.ICON_QUESTION)
            abortFlag = False
            if (confdlg.ShowModal() == wx.ID_OK):
              shutil.copyfile(os.path.join(dirname, filename), os.path.join(workdir, filename))
            else:
              abortFlag = True

            confdlg.Destroy()
            if(abortFlag == True):
              if(utilities.verifyKeyFile(os.path.join(workdir, currentList.PvtKeyFile), filetype, currentList) == False):
                # don't change
                self.pvtKeyFileEdit.ChangeValue("")
                currentList.PvtKeyFile = ""
              self.StatusBar.SetStatusText( "Copy cancelled, using exiting key" )
          else:
            shutil.copyfile(os.path.join(dirname, filename), os.path.join(workdir, filename))

    dlg.Destroy()
    #print("onPvtBrowseButtonClick: %s" % (currentList.PvtKeyFile))   # DBGDBG

  def onPubBrowseButtonClick(self, event):
    """ onPubBrowseButtonClick - browse to an existing PDEF file"""

    #self.StatusBar.SetStatusText("You clicked the Browse button")
    dirname = ''   #  current working directory
    workdir = self.pdef.WorkingDirectory
    wildcard = "Key file (*.pem) | *.pem|" \
               "All Files (*.*)    | *.*"
    dlg = wx.FileDialog(self.parent, "Choose the Public Key file", workdir, "", wildcard, wx.FD_OPEN)

    currentList = self.pdef.getCurrentListObject()
    if currentList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']:
      filetype = DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']
    else:
      # treat SM2 as ECDSA key
      filetype = DEFINES.KEY_FILE_TYPE['PUBLIC_ECDSA']

    if dlg.ShowModal() == wx.ID_OK:
      filename = dlg.GetFilename()
      dirname  = dlg.GetDirectory()

      self.pubKeyFileEdit.ChangeValue(filename)
      currentList.PubKeyFile = filename
      self.setListModified()
      if(utilities.verifyKeyFile(os.path.join(dirname, currentList.PubKeyFile), filetype, currentList) == False):
        self.pubKeyFileEdit.ChangeValue("")
        currentList.PubKeyFile = ""
      elif (dirname != workdir):
        if (os.path.exists(os.path.join(workdir, filename))) :
          confdlg = wx.MessageDialog(self.parent, filename+" already exists in working directory\nOverwrite file in working directory?", "Confirm Copy", wx.OK|wx.CANCEL|wx.ICON_QUESTION)
          abortFlag = False
          if (confdlg.ShowModal() == wx.ID_OK):
            shutil.copyfile(os.path.join(dirname, filename), os.path.join(workdir, filename))
          else:
            abortFlag = True

          confdlg.Destroy()
          if(abortFlag == True):
            # verify the existing file
            if(utilities.verifyKeyFile(os.path.join(workdir, currentList.PubKeyFile), filetype, currentList) == False):
              self.pubKeyFileEdit.ChangeValue("")
              currentList.PubKeyFile = ""
            self.StatusBar.SetStatusText( "Copy cancelled, using exiting key" )
        else:
          shutil.copyfile(os.path.join(dirname, filename), os.path.join(workdir, filename))

    dlg.Destroy()
    #print("onPubBrowseButtonClick: %s" % (currentList.PubKeyFile))   # DBGDBG

  def onAlgorithmEdit(self, event):
    """ onAlgorithmEdit - select the algorithm"""

    func = "onAlgorithmEdit"
    currentList = self.pdef.getCurrentListObject()
    selectionString = event.GetString()
    algName = selectionString.split('/')
    if len(algName) == 2:
      sigAlg = algName[0].split(' ')[0]
      hashAlg = algName[1]

      # Check for valid signature algorithm name
      try:
        sigAlgName = (key for key in DEFINES.TPM_ALG_SIGN.keys() if (sigAlg in key)).next()
      except StopIteration:
        self.StatusBar.SetStatusText("Signature algorithm name mismatch")
        print("%s - Signature algorithm name mismatch" % (func))  # DBGDBG
        return

      # Check for valid algorithm name
      try:
        hashAlgName = (key for key in DEFINES.TPM_ALG_HASH.keys() if (hashAlg in key)).next()
      except StopIteration:
        self.StatusBar.SetStatusText("Hash algotithm name mismatch")
        print("%s - Hash algorithm name mismatch" % (func))  # DBGDBG
        return

      currentList.SigAlgorithm = DEFINES.TPM_ALG_SIGN[sigAlgName]
      currentList.sigAlgorithmHash = DEFINES.TPM_ALG_HASH[hashAlgName]
      if currentList.KeySize not in DEFINES.SIGNATURE_KEY_SIZE[selectionString]:
        currentList.KeySize = DEFINES.SIGNATURE_KEY_SIZE[selectionString][0]
      self.enableDisableSigned(True, selectionString)
    else:
      currentList.SigAlgorithm = DEFINES.TPM_ALG_SIGN['NULL']
      self.enableDisableSigned(False)


  def onAddElementClick(self, event):
    """onAddElementClick"""
    func = "onAddElementClick"

    # ask user to select the type of element and its hash algorithm
    if(self.pdef.Rules == DEFINES.PsRules):
      addElementChoices = self.addElementChoicesPs
    elif(self.pdef.Rules == DEFINES.PoRules):
      addElementChoices = self.addElementChoicesPo
    else:
      print("%s - invalid pdef.Rules=%d" % (func, self.pdef.Rules))
      return

    message = "Select the type of element and its hash algorithm"
    title   = "Add Element Selections"
    dialog = wx.SingleChoiceDialog(None, message, title, addElementChoices)
    if dialog.ShowModal() == wx.ID_OK:
      requestedElement = dialog.GetStringSelection()
    else:
      dialog.Destroy()
      return          # cancel out

    dialog.Destroy()
    self.StatusBar.SetStatusText("Adding a %s element to this LIST" % (requestedElement))

    print("%s: adding %s" % (func, requestedElement))  # DBGDBG
    self.deleteElementButton.Enable( True )
    currentListObject = self.pdef.getCurrentListObject()
    self.selectElementEdit.Enable( True )
    currentElementsCnt = int(self.numberElementsEdit.GetValue())
    currentElementsCnt += 1
    self.numberElementsEdit.ChangeValue(str(currentElementsCnt))

    # if an element is being shown, hide it.  Then show the new one
    if(currentListObject.CurrentElementView != DEFINES.ELEMENT_NAME_NONE):
      self.hideThisPanel(currentListObject.CurrentElementView)

    currentListObject.CurrentElementView = requestedElement
    self.selectElementEdit.Append(requestedElement)
    currentSelection = self.selectElementEdit.GetSelection()
    #print("%s: currentSelection=%d newSelection=%d" % (func, currentSelection, currentSelection+1)) # DBGDBG
    self.selectElementEdit.SetSelection(currentSelection + 1)
    self.selectElementEdit.SetValue(requestedElement)
    self.updateAddElementChoicesAfterAdd(requestedElement)
    self.visibleElement = requestedElement

    # Create a new element of selected spec and populate into currentListObject.MleDefData
    elementType, hashAlg = requestedElement.split('-')

    if hashAlg == 'LEGACY':
      if elementType == 'SBIOS':
        element = SBIOSLegacy()
        defdata = SBIOSLEGACY_DEF()
      elif elementType == 'MLE':
        element = MLELegacy()
        defdata = MLELEGACY_DEF()
      elif elementType == 'PCONF':
        element = PCONFLegacy()
        defdata = PCONFLEGACY_DEF()
      else:
        print ("ERROR: invalid element")
    else:
      if elementType == 'SBIOS':
        element = SBIOS(DEFINES.TPM_ALG_HASH[hashAlg])      # GUI panel
        defdata = SBIOS_DEF(DEFINES.TPM_ALG_HASH[hashAlg])  # Element defined in pdef.py
      elif elementType == 'STM':
        element = STM(DEFINES.TPM_ALG_HASH[hashAlg])
        defdata = STM_DEF(DEFINES.TPM_ALG_HASH[hashAlg])
      elif elementType == 'MLE':
        element = MLE(DEFINES.TPM_ALG_HASH[hashAlg])
        defdata = MLE_DEF(DEFINES.TPM_ALG_HASH[hashAlg])
      elif elementType == 'PCONF':
        element = PCONF(DEFINES.TPM_ALG_HASH[hashAlg])
        defdata = PCONF_DEF(DEFINES.TPM_ALG_HASH[hashAlg])
      else:
        print ("ERROR: invalid element")

    currentListObject.ElementDefData.append(defdata)
    element.createOrShowPanel(wx, self, self.parent, self.pdef, self.StatusBar)
    self.includedElements.append(element)

    # Note - panels were constructed with the correct hashAlg, so no need to pass that thru here
    # Note - iterators not used here since need to update the pdef's  xxxDefData.IncludeXXX not the array's

    self.selectElementEdit.Enable(True)
    self.parent.Layout()


  def syncVersion(self, rule):
    currentList = self.pdef.getCurrentListObject()
    listversion = str(currentList.ListVersionMajor)+'.'+str(currentList.ListVersionMinor)
    self.versionEdit.ChangeValue(listversion)
    if rule == DEFINES.PoRules:
      enable = True
    else:
      enable = False
    for element in self.includedElements:
      element.restorePanel(currentList, self.pdef.MaxHashes)
      element.enableDisableOverridePsPolicy(enable)


  def checkListModified(self, pdef):
    """checkListModified - return True if any list has been modified, else False"""

    # pdef.PolListInfo is a dictionary of 8 PLIST_DEF's from '0' to '7'
    # or None if that list was not added
    # i.e. {'0':PLIST_DEF or None, ... '7':PLIST_DEF or None}
    #
    i = '0'
    while(i<'8'):
      #print("checkListModified list %s" % (i))  # DBGDBG
      if(pdef.PolListInfo[i] != None):              # if list exists
        if(pdef.PolListInfo[i].ListModified == True):
          print("checkListModified Return - list %s was modified" % (i))  # DBGDBG
          return True

      i = str(int(i) + 1)

    return False


  def setListModified(self):
    """setListModified - if list not modified yet, increment its rev cnt and set it to modified"""

    currentList = self.pdef.getCurrentListObject()
    #also set pdef.Modified for saving file\
    self.pdef.Modified =  True
    #print("setListModified - ListModified=%s" % (currentList.ListModified))  # DBGDBG
    if(currentList.ListModified == False):
      currentList.RevocationCounter += 1
      self.resetButton.Enable(True)

      self.revocationCountEdit.ChangeValue(str(currentList.RevocationCounter))   # update the GUI
      currentList.ListModified = True

  def rebuildSelectElementChoices(self):
    """rebuildSelectElementChoices - rebuild selectElementEdit's choices list for the elements existing in this view"""
    func = 'rebuildSelectElementChoices'

    self.selectElementEdit.Clear()
    self.selectElementEdit.Append('None')
    # For each existing element in this list (ie IncludeXXXX==True)
    currentList = self.pdef.getCurrentListObject()
    for element in currentList.ElementDefData:
      self.selectElementEdit.Append(element.Name)


  # remove the added element from the add element selection list
  def updateAddElementChoicesAfterAdd(self, requestedElement):
    """updateAddElementChoicesAfterAdd"""
    func = 'updateAddElementChoicesAfterAdd'

    if(self.pdef.Rules == DEFINES.PoRules):
      self.addElementChoicesPo.remove(requestedElement)
      if(len(self.addElementChoicesPo) == 0):
        self.addElementButton.Enable(False)
    if(self.pdef.Rules == DEFINES.PsRules):
      self.addElementChoicesPs.remove(requestedElement)
      if(len(self.addElementChoicesPs) == 0):
        self.addElementButton.Enable(False)

  # append the deleted element to the add element selection list
  def updateAddElementChoicesAfterDelete(self, deletedElement):
    """updateAddElementChoicesAfterDelete"""
    func='updateAddElementChoicesAfterDelete'

    # append the deleted element to the add element selection list
    self.addElementChoicesPs.append(deletedElement)
    self.addElementChoicesPo.append(deletedElement)


  # rebuild the add element selection list to include all the elements that don't exist
  def rebuildAddElementChoices(self):
    """rebuildAddElementChoices"""
    self.rebuildAddElementChoicesForPoRules()
    self.rebuildAddElementChoicesForPsRules()

  # if a PO rules element doesn't exist, add it to the selections list
  def rebuildAddElementChoicesForPoRules(self):
    """rebuildAddElementChoicesForPoRules"""
    func='rebuildAddElementChoicesForPoRules'

    #Element = DEFINES.ELEMENT_PO_RULES
    self.addElementChoicesPo = []
    self.addElementChoicesPo += DEFINES.ELEMENT_PO_RULES
    currentList = self.pdef.getCurrentListObject()

    for element in currentList.ElementDefData:
      if element.Name in self.addElementChoicesPo:
        self.addElementChoicesPo.remove(element.Name)
      else:
        print ("Element Name %s not found in Elements for PO rule" %(element.Name))

    if len(currentList.ElementDefData) >= 8 or len(self.addElementChoicesPs) == 0:
      flag = False
    else:
      flag = True
    self.addElementButton.Enable(flag)
    #print("%s - addElementChoicesPo=%s" % (func, self.addElementChoicesPo)) #DBGDBG

  # if a PS rules element doesn't exist, add it to the selections list
  def rebuildAddElementChoicesForPsRules(self):
    """rebuildAddElementChoicesForPsRules"""
    func='rebuildAddElementChoicesForPsRules'

    #Element = DEFINES.ELEMENT                    # gets all the elements i.e. = PS rules
    self.addElementChoicesPs = []
    self.addElementChoicesPs += DEFINES.ELEMENT   # gets all the elements i.e. = PS rules
    currentList = self.pdef.getCurrentListObject()

    for element in currentList.ElementDefData:
      if element.Name in self.addElementChoicesPs:
        self.addElementChoicesPs.remove(element.Name)
      else:
        print ("Element Name %s not found in Elements for PS rule" %(element.Name))

    if len(currentList.ElementDefData) >= 8 or len(self.addElementChoicesPs) == 0:
      flag = False
    else:
      flag = True
    self.addElementButton.Enable(flag)
    #print("%s - addElementChoicesPs=%s" % (func, self.addElementChoicesPs)) #DBGDBG

  def loadAndDisplayPlistDefFile(self, file, wx, pdef, statusBar, parent):
    """ loadAndDisplayPlistDefFile - get a saved plistDef object from a PDEF file and update the list panel"""
    func = 'loadAndDisplayPlistDefFile'

    self.pdef = pdef
    self.StatusBar = statusBar
    self.parent = parent

    print("%s load plistDef %d, NumLists=%d" % (func, pdef.CurrentListView, pdef.NumLists))     # DBGDBG
    currentListObject = self.pdef.getCurrentListObject()


    # If this list is to be the displayed list panel[ie pdef.CurrentListView]
    # then restore the list panel & enable/disable the AddElement buttons
    # restore the list's element panels that exist but hide them
    # only show & restore the specified element [ie list.CurrentElementView]
    #
    print("%s thisListNum=%d, CurrentListView=%d" % (func, pdef.CurrentListView, pdef.CurrentListView))     # DBGDBG
    if(self.listPanelWidgets == []):
      self.createListPanel(wx, parent, pdef.CurrentListView, pdef, statusBar)
    else:
      # in case list or element panels all ready exist ...
      self.hideListPanel()
      for i in self.listPanelWidgets:
        i.Show()

    self.setListPanelToCurrentListView()
    #currentListObject.ListModified = False

  def printPlistDefs(self, pdef, f):
    """printPlistDefs - write all created PLIST_DEFs to the specified human readable text file for printing"""

    # pdef.PolListInfo is a dictionary of 8 PLIST_DEF's from '0' to '7'
    # or None if that list was not added
    # i.e. {'0':PLIST_DEF or None, ... '7':PLIST_DEF or None}
    #
    i = '0'             # i is 0 based index for loop control
    cnt = '1'         # cnt is 1 based 'i' for printing
    while(i<'8'):
      print("PlistInfo", cnt, " = ", pdef.PolListInfo[i], file=f)
      if(pdef.PolListInfo[i] != None):
        self.printPlistDef(pdef.PolListInfo[i], f)     # write this PolListInfo[]

      i = str(int(i) + 1)
      cnt = str(int(cnt) + 1)

  def printPlistDef(self, plistDef, f):
    """printPlistDef - write this PLIST_DEF to the specified human readable text file for printing"""

    #print("printPlistDef - PLIST_DEF object: %s" % (plistDef))     # DBGDBG

    print("LdefSize",             " = ", plistDef.LdefSize,            file=f)
    print("Tag",                  " = ", plistDef.Tag,                 file=f)
    print("ListVersion",          " = ", plistDef.ListVersionMajor, ".", plistDef.ListVersionMinor, sep='', file=f)
    print("ListValid",            " = ", plistDef.ListValid,           file=f)
    print("ListModified",         " = ", plistDef.ListModified,        file=f)
    print("SigAlgorithm",         " = ", hex(plistDef.SigAlgorithm),        file=f)
    print("sigAlgorithmHash",     " = ", hex(plistDef.sigAlgorithmHash),    file=f)
    print("PolicyElementSize",    " = ", plistDef.PolicyElementSize,   file=f)
    print("CurrentElementView",   " = ", plistDef.CurrentElementView,  file=f)
    print("SyncRevCount",         " = ", plistDef.SyncRevCount,        file=f)
    print("AllowedCounter",       " = ", plistDef.RevokeCounter,       file=f)
    print("RevocationCounter",    " = ", plistDef.RevocationCounter,   file=f)
    print("KeySize",              " = ", plistDef.KeySize,             file=f)
    print("PubKeyFile",           " = ", plistDef.PubKeyFile,          file=f)
    print("PvtKeyFile",           " = ", plistDef.PvtKeyFile,          file=f)

    currentListObject = plistDef

    # print the element summary showing which elements were created
    # Lists always contain all elements, even if they haven't been added
    # so need to check if they should be included
    for element in plistDef.ElementDefData:
      print("%s DefData[Index]"% (element.Name),  " = ", element,  file=f)
      element.printDef(f)
      print("\n", file=f)         # for readability


#  def enablePconfOverridePsPolicyCheckbox(self, value):
#    """enablePconfOverridePsPolicyCheckbox - if current list has a PCONF element, enable its OverridePsPolicy checkbox"""
#
#    currentList = self.pdef.getCurrentListObject()
#    if(currentList.PconfDefData[DEFINES.DEFDATA_INDEX['SHA256']].IncludeInList == True):
#      pconf256.enablePconfOverridePsPolicyCheckbox(value)
#    if(currentList.PconfDefData[DEFINES.DEFDATA_INDEX['SHA1']].IncludeInList == True):
#      pconf1.enablePconfOverridePsPolicyCheckbox(value)
#    if(currentList.PconfLegacyDefData[DEFINES.DEFDATA_INDEX['SHA1']].IncludeInList == True):
#      pconf0.enablePconfOverridePsPolicyCheckbox(value)


  def hideAllPanels( self ):
    """hideAllPanels - hide all the element panels"""

    for element in self.includedElements:
      element.hidePanel()


  def hideThisPanel( self, panelToHide ):
    """hideThisPanel"""

    for element in self.includedElements:
      found = element.isElementType(panelToHide)
      if found:
        element.hidePanel()

    print("hideThisPanel: %s" % (panelToHide)) # DBGDBG

  # the last function in the file doesn't show up in the scope list in Understand for some reason!
  def stub(self):
    pass

